/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamturntable.h"
#include <stdlib.h>
#include <gdk/gdk.h>
#include <gtk/gtksignal.h>
#include <gdam/gdamsample.h>
#include <gdam/gdamrequest.h>
#include "gdamutil.h"
#include <math.h>
#include <string.h>

/*
 * Strategy:
 *       producing->beat_info the beat information for the underlying
 *                            producing model.
 *       this->beat_info      the resampled beat info.
 */

static void gdam_turntable_add_xml  (GdamModel*        model,
                                     xmlNode*          parent,
				     GdamXmlInfo*      xml_info);
static void gdam_turntable_parse_xml(GdamModel*        model,
                                     xmlNode*          node,
				     GdamXmlInfo*      xml_info);

/*   ____ _                 ___       _ _   
 *  / ___| | __ _ ___ ___  |_ _|_ __ (_) |_ 
 * | |   | |/ _` / __/ __|  | || '_ \| | __|
 * | |___| | (_| \__ \__ \  | || | | | | |_ 
 *  \____|_|\__,_|___/___/ |___|_| |_|_|\__|
 */                                        


#define PARTIAL 1

enum {
	UPDATE_ANCHORED,
	UPDATE_BEATNUM,
	SET_RATE,
	UPDATE_SYNC_PREVIEW,
	UPDATE_JUMP_PREVIEW,
	LAST_SIGNAL
};

static guint turntable_signals[LAST_SIGNAL];

/* Model class functions. */
static gboolean gdam_turntable_match_beat_info(GdamModel* model,
					       const GdamBeatInfo* match_info);
/* Producer class functions */
static void     gdam_turntable_play                (GdamProducer*  turntable,
						    GdamPlayInfo*  play_info);
static void     gdam_turntable_real_stop           (GdamProducer*  turntable);
static void     gdam_turntable_real_set_position   (GdamProducer*  producer,
					            gdouble	   position);
static void 	gdam_turntable_real_set_rate	   (GdamTurntable* turntable, 
						    gdouble 	   newrate);
#if 0
static void     handle_producer_set_source_id      (GdamProducer*  producer,
			                            int       	   new_id,
			                            GdamTurntable* turntable);
#endif
static void     gdam_turntable_new_request         (GdamModel* turntable,
			                            GdamModel* producer,
						    GdamPlayInfo*  play_info,
			                            int            sub_id,
			                            GdamRequestBuilder* b);

static void _gdam_turntable_class_init(GdamTurntableClass* klass)
{
	GdamModelClass* model_class = GDAM_MODEL_CLASS(klass);
	GdamProducerClass* producer_class = GDAM_PRODUCER_CLASS(klass);
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(klass);
	model_class->match_beat_info = gdam_turntable_match_beat_info;
	model_class->new_request = gdam_turntable_new_request;
	model_class->add_xml = gdam_turntable_add_xml;
	model_class->parse_xml = gdam_turntable_parse_xml;
	producer_class->stop = gdam_turntable_real_stop;
	producer_class->play = gdam_turntable_play;
        producer_class->set_position = gdam_turntable_real_set_position;
	klass->set_rate = gdam_turntable_real_set_rate;


	turntable_signals[UPDATE_BEATNUM] =
		gtk_signal_new(
			"update_beatnum",
			GTK_RUN_LAST,
			object_class->type,
			GTK_SIGNAL_OFFSET(GdamTurntableClass, update_beatnum),
			gtk_marshal_NONE__DOUBLE,
			GTK_TYPE_NONE,
			1,
			GTK_TYPE_DOUBLE
		);
	turntable_signals[UPDATE_ANCHORED] =
		gtk_signal_new(
			"update_anchored",
			GTK_RUN_LAST,
			object_class->type,
			GTK_SIGNAL_OFFSET(GdamTurntableClass, update_anchored),
			gtk_marshal_NONE__INT,
			GTK_TYPE_NONE,
			1,
			GTK_TYPE_INT
		);
	turntable_signals[SET_RATE] =
		gtk_signal_new(
			"set_rate",
			GTK_RUN_LAST,
			object_class->type,
			GTK_SIGNAL_OFFSET(GdamTurntableClass, set_rate),
			gtk_marshal_NONE__DOUBLE,
			GTK_TYPE_NONE,
			1,
			GTK_TYPE_DOUBLE
		);
	turntable_signals[UPDATE_SYNC_PREVIEW] =
		gtk_signal_new(
			"update_sync_preview",
			GTK_RUN_LAST,
			object_class->type,
			GTK_SIGNAL_OFFSET(GdamTurntableClass, 
							update_sync_preview),
			gtk_marshal_NONE__DOUBLE,
			GTK_TYPE_NONE,
			1,
			GTK_TYPE_DOUBLE
		);
	turntable_signals[UPDATE_JUMP_PREVIEW] =
		gtk_signal_new(
			"update_jump_preview",
			GTK_RUN_LAST,
			object_class->type,
			GTK_SIGNAL_OFFSET(GdamTurntableClass, 
							update_jump_preview),
			gtk_marshal_NONE__DOUBLE,
			GTK_TYPE_NONE,
			1,
			GTK_TYPE_DOUBLE
		);
	gtk_object_class_add_signals(object_class, 
				     turntable_signals,
				     LAST_SIGNAL);
}

static void _gdam_turntable_init(GdamTurntable* turntable)
{
	turntable->gain = 1.0;
	turntable->rate = 1.0;
	turntable->bpm_factor = 1.0;
}

GtkType gdam_turntable_get_type()
{
	static GtkType turntable_type = 0;
	if (!turntable_type) {
		static const GtkTypeInfo turntable_info = {
			"GdamTurntable",
			sizeof(GdamTurntable),
			sizeof(GdamTurntableClass),
			(GtkClassInitFunc) _gdam_turntable_class_init,
			(GtkObjectInitFunc) _gdam_turntable_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		turntable_type = gtk_type_unique(GDAM_TYPE_PRODUCER,
		                                 &turntable_info);
	}
	return turntable_type;
}

GdamTurntable* gdam_turntable_new    (GdamChannel*        channel) 
{
	GdamTurntable* turntable;
	turntable = GDAM_TURNTABLE(gtk_type_new(GDAM_TYPE_TURNTABLE));
	((GdamModel*)turntable)->channel = channel;
	return turntable;
}

/*
 *            _ _ _                _
 *   ___ __ _| | | |__   __ _  ___| | _____
 *  / __/ _` | | | '_ \ / _` |/ __| |/ / __|
 * | (_| (_| | | | |_) | (_| | (__|   <\__ \
 *  \___\__,_|_|_|_.__/ \__,_|\___|_|\_\___/
 * 
 */

/* TODO: scale stopped_pos, length */
void resample_rate_changed(GdamTurntable* turntable)
{
	GdamProducer* producer = turntable->producer;
	GdamModel* model = GDAM_MODEL(turntable);
	GdamModel* producer_model = GDAM_MODEL(producer);
	GdamChannel* channel = model->channel;

	if (turntable->resample_id){
		GdamArg arg;
		GdamRequest* request;
		arg.name = "rate";
		arg.type = GDAM_ARG_TYPE_DOUBLE;
		GDAM_ARG_DOUBLE(&arg) = turntable->rate;
		request = gdam_request_set_args(
				turntable->resample_id,
				1,
				&arg,
				gdam_do_nothing,
				gdam_do_nothing,
				NULL);
		gdam_channel_append_request(channel, request);
	}
	if ( producer != NULL) {
		GdamBeatInfo prod_beat = producer_model->beat_info;
		GdamBeatInfo new_beat;
		float new_bpm;
		float new_first_beat;
		new_bpm = gdam_beat_info_get_bpm(&prod_beat, 0)
				* turntable->rate / turntable->bpm_factor;
		new_first_beat = gdam_beat_info_get_first_beat(&prod_beat, 0)
						/ turntable->rate;
		gdam_beat_info_set_bpm(&new_beat, new_bpm);
		gdam_beat_info_set_first_beat(&new_beat, new_first_beat);
		gdam_model_set_beat_info(model, &new_beat);
		if (producer_model->source_id != 0
	         && !producer_model->create_source_pending) {
			float stop_pos;
			stop_pos = producer->stopped_position / turntable->rate;
			gdam_producer_set_position(GDAM_PRODUCER(turntable), 
						   stop_pos);
		}
	}
	
	return;
}

static void sync_preview_success(GdamTime preview_time, gpointer data)
{
	GdamTurntable* turntable = GDAM_TURNTABLE(data);
	gtk_signal_emit(GTK_OBJECT(turntable),
			turntable_signals[UPDATE_SYNC_PREVIEW],
			(double)preview_time / 44100.0);
	gtk_object_unref(GTK_OBJECT(turntable));
}

static void sync_preview_failure(gpointer data)
{
	GdamTurntable* turntable = GDAM_TURNTABLE(data);
	gtk_object_unref(GTK_OBJECT(turntable));
}

void turntable_do_sync_preview(GdamTurntable* turntable)
{
	GdamBeatInfo new_beat_info;
	guint leader_id;
	GdamModel* model = (GdamModel*)turntable;
	//GdamModel* model = (GdamModel*)(turntable->producer);  
	GdamRequest* request;
	float comp_bpm;
	float comp_first_beat;
	float lead_bpm;
	float lead_first_beat;
	g_return_if_fail(model != NULL);
	if (turntable == NULL){
		g_warning("missing needed pointers in do_sync_preview");
	}
	if (!model->source_id)
		return;
	if (turntable->anchored) {
		/*
		gtk_signal_emit(GTK_OBJECT(turntable),
				turntable_signals[UPDATE_SYNC_PREVIEW],
				0.0);
		*/
		return;
	}
	if (!gdam_infomanager_get_sync_info(model,
					    &leader_id,
					    &new_beat_info))
		return;
	comp_bpm = gdam_beat_info_get_bpm(&model->beat_info, 0);
	comp_first_beat = gdam_beat_info_get_first_beat(&model->beat_info,0);
	lead_bpm = gdam_beat_info_get_bpm(&new_beat_info, 0);
	lead_first_beat = gdam_beat_info_get_first_beat(&new_beat_info,0);
	request = gdam_request_sync_preview(
				       model->source_id,
				       leader_id,
				       comp_bpm,
				       comp_first_beat,
				       lead_bpm,
				       lead_first_beat,
				       sync_preview_success,
				       sync_preview_failure,
				       turntable);
	gdam_channel_append_request(model->channel, request);
	gtk_object_ref(GTK_OBJECT(turntable));
	return;
}

void turntable_do_sync(GdamTurntable* turntable)
{
	GdamRequest* request;
	GdamBeatInfo static_beat_info;
	GdamModel* model = GDAM_MODEL(turntable);
	guint leader_id;
	float comp_bpm;
	float comp_first_beat;
	float static_bpm;
	float static_first_beat;
	if (turntable == NULL
	 || model->source_id == 0)
		return;
	if (turntable->anchored) {
		gtk_signal_emit(GTK_OBJECT(turntable),
				turntable_signals[UPDATE_SYNC_PREVIEW],
				0.0);
		return;
	}
	if (!gdam_infomanager_get_sync_info(model,
					    &leader_id,
					    &static_beat_info))
		return;
	comp_bpm = gdam_beat_info_get_bpm(&model->beat_info,0),
	comp_first_beat =  gdam_beat_info_get_first_beat(&model->beat_info,0);
	static_bpm =  gdam_beat_info_get_bpm(&static_beat_info,0),
	static_first_beat = gdam_beat_info_get_first_beat(&static_beat_info,0);
	gdam_debug("sync: mobile id: %d bpm: %f fb: %f",
	          model->source_id,
		  comp_bpm,
		  comp_first_beat);
	gdam_debug("sync: static id: %d bpm: %f fb: %f",
	          leader_id,
		  static_bpm,
		  static_first_beat);
	request = gdam_request_synchronize (GDAM_MODEL(turntable)->source_id,
	                                    leader_id, 
					    comp_bpm,
					    comp_first_beat,
					    static_bpm,
					    static_first_beat,
					    gdam_do_nothing,
					    gdam_do_nothing,
					    NULL);
	gdam_channel_append_request(model->channel, request);
	gtk_signal_emit(GTK_OBJECT(turntable),
			turntable_signals[UPDATE_SYNC_PREVIEW],
			0.0);
}


void gdam_turntable_seek_by_beats(GdamTurntable* turntable, gdouble jumpAmt)
{
	gdouble bpm;
	/* TODO: check for unreasonable values? */
	gdam_debug("tt:seek_by_beats - %f", jumpAmt);
	if (!turntable->producer){
        	gdam_debug("no producer, aborting seek_by_beats");
		return;
	}
	bpm = gdam_beat_info_get_bpm(
			&GDAM_MODEL(turntable->producer)->beat_info, 0);
    	if (bpm < 1.0) 
		gdam_turntable_seek(turntable, jumpAmt * 1.0);
	else{
		float length;
		length = jumpAmt * 60.0 / bpm;
		if (turntable->bpm_factor > 0)
			length *= turntable->bpm_factor;
		gdam_turntable_seek(turntable, length);
	}
}

void gdam_turntable_seek(GdamTurntable* turntable, gdouble jump_amount)
{
	gdam_debug("tt:seek - %f", jump_amount);
	if (turntable->producer 
		&& GDAM_MODEL(turntable->producer)->source_id) {
		gdam_producer_seek(turntable->producer, jump_amount, 1);
	} else {
		gdouble new_pos = GDAM_PRODUCER(turntable)->stopped_position;
		new_pos += jump_amount / turntable->rate;
		gdam_producer_set_position(GDAM_PRODUCER(turntable), new_pos);
	}
}

static void gdam_turntable_real_stop(GdamProducer* producer)
{
	GdamChannel* channel = ((GdamModel*)producer)->channel;
	GdamTurntable* turntable = GDAM_TURNTABLE(producer);

	gdam_debug("turntable_stop called");

	if (turntable->pos_sub_id){
		gdam_channel_unsubscribe(channel, turntable->pos_sub_id);
		turntable->pos_sub_id = 0;
	}
	if (turntable->eof_sub_id){
		gdam_channel_unsubscribe(channel, turntable->eof_sub_id);
		turntable->eof_sub_id = 0;
	}
	if (!turntable->producer){
		gdam_debug("turntable_stop aborting, no producer.");
		return;
	}
	gdam_producer_stop(turntable->producer);
}

static void producer_ended_callback(gpointer user_data)
{
	GdamTurntable* turntable = GDAM_TURNTABLE(user_data);
	/* TODO: emit stopped signal */
	gdam_turntable_real_stop(GDAM_PRODUCER(turntable));
}

static void producer_position_callback(GdamTime time, gpointer user_data)
{
	GdamTurntable* turntable = GDAM_TURNTABLE(user_data);
	float pos = (double)time / 44100.0;
	pos /= turntable->rate;
	//gdam_debug("turntable_producer_pos_callback: %f .", pos);
	if (!turntable->user_dragging_position)
		gdam_producer_set_position(GDAM_PRODUCER(turntable), pos);
}


static void on_resample_create_func(int resample_id, gpointer turntable_ptr)
{
       GdamTurntable* turntable = GDAM_TURNTABLE(turntable_ptr);
       turntable->resample_id = resample_id;
       gdam_debug("gdam_turntable: resample created successfully, id %d",
                       resample_id);
}

static void on_resample_create_fail_func(gpointer turntable_ptr)
{
       GdamTurntable* turntable = GDAM_TURNTABLE(turntable_ptr);
       gdam_debug("gdam_turntable [%p]: resample created failed", turntable);
}

static GdamRequest*
build_resample_request(GdamTurntable* turntable, guint subsource_id)
{
	return gdam_request_create_source(
				"GdamResample",
				on_resample_create_func,
				on_resample_create_fail_func,
				turntable,
				2,
				"rate",
				GDAM_ARG_TYPE_DOUBLE,
				turntable->rate,
				"filtered",
				GDAM_ARG_TYPE_INT,
				subsource_id);
}

static void on_gain_create_func(int gain_id, gpointer turntable_ptr)
{
	GdamTurntable* turntable = GDAM_TURNTABLE(turntable_ptr);
	gdam_model_set_source_id(GDAM_MODEL(turntable), gain_id);
	gdam_debug("gdam_turntable: gain created successfully, id %d",
			gain_id);
}

static void on_gain_create_fail_func(gpointer turntable_ptr)
{
	GdamTurntable* turntable = GDAM_TURNTABLE(turntable_ptr);
	gdam_debug("gdam_turntable [%p]: gain created failed", turntable);
}


static GdamRequest*
build_gain_request(GdamTurntable* turntable, guint tmp_resample_id)
{
	return gdam_request_create_source(
				"GdamGain",
				on_gain_create_func,
				on_gain_create_fail_func,
				turntable,
				2,
				"gain",
				GDAM_ARG_TYPE_DOUBLE,
				turntable->gain,
				"filtered",
				GDAM_ARG_TYPE_INT,
				tmp_resample_id);
}

static void     handle_producer_set_beat_info      (GdamProducer*  producer,
			                            const GdamBeatInfo* bi,
			                            GdamTurntable* turntable)
{
	GdamBeatInfo beat_info;
	float bpm = gdam_beat_info_get_bpm(bi, 0);
	float first_beat = gdam_beat_info_get_first_beat(bi, 0);
	bpm = gdam_beat_info_get_bpm(bi,0) * turntable->rate 
						/ turntable->bpm_factor;
	first_beat = gdam_beat_info_get_first_beat(bi,0) / turntable->rate;
	gdam_beat_info_set_bpm(&beat_info, bpm);
	gdam_beat_info_set_first_beat(&beat_info, first_beat);
	gdam_debug("handle_producer_set_beat_info called");
	gdam_model_set_beat_info(GDAM_MODEL(turntable), &beat_info);
}

static void     handle_producer_set_position       (GdamProducer*  producer,
			                            double	   prod_pos,
			                            GdamTurntable* turntable)
{
	float tt_position;
	tt_position = prod_pos / turntable->rate; 
	gdam_producer_set_position(GDAM_PRODUCER(turntable), tt_position);
}

static void     handle_producer_update_length      (GdamProducer*  producer,
			                            double	   prod_len,
			                            GdamTurntable* turntable)
{
	float tt_len;
	tt_len = prod_len / turntable->rate; 
	gdam_producer_update_length(GDAM_PRODUCER(turntable), tt_len);
}

static void append_unsubscribe_events(GdamTurntable* turntable,  
                                      GdamRequestBuilder* builder)
{
	if (turntable->eof_sub_id) {
		gdam_request_builder_append_generic(
				builder,
				gdam_request_unsubscribe(
					 turntable->eof_sub_id,
					 NULL,
					 NULL,
					 NULL
				)
		);
		turntable->eof_sub_id = 0;
	}
	if (turntable->pos_sub_id) {
		gdam_request_builder_append_generic(
				builder,
				gdam_request_unsubscribe(
					 turntable->pos_sub_id,
					 NULL,
					 NULL,
					 NULL
				)
		);
		turntable->pos_sub_id = 0;
	}
}


static void eof_connect_success(int eof_sub_id, gpointer user_data)
{
	GdamTurntable* turntable = GDAM_TURNTABLE(user_data);
	turntable->eof_sub_id = eof_sub_id;
	gdam_channel_register_handler(
		GDAM_MODEL(turntable)->channel,
		gdam_channel_simple_handler_new(eof_sub_id,
						TRUE,
					        producer_ended_callback,
						turntable)
	);
}

static void 
position_callback_connect_success(int pos_sub_id, gpointer user_data)
{
	GdamTurntable* turntable = GDAM_TURNTABLE(user_data);
	turntable->pos_sub_id = pos_sub_id;
	gdam_channel_register_handler(
		GDAM_MODEL(turntable)->channel,
		gdam_channel_time_handler_new(pos_sub_id,
					      FALSE,
					      producer_position_callback,
					      turntable)
	);
}

static void append_subscribe_events(GdamTurntable*      turntable,
                                    GdamRequestBuilder* builder,
				    int                 sub_source_id)
{
	g_return_if_fail(turntable->producer != NULL);
	gdam_request_builder_append_generic(
			builder,
			gdam_request_subscribe_source_ended(
				 sub_source_id,
				 eof_connect_success,
				 NULL,
				 turntable
			)
	);

	gdam_request_builder_append_generic(
			builder,
			gdam_request_subscribe_source_timer(
				 sub_source_id,
				 200,
				 position_callback_connect_success,
				 NULL,
				 turntable
			)
	);
	gdam_debug("turntable: appending subscribe_timer request");
}

static void gdam_turntable_new_request(GdamModel*           turntable_model,
			               GdamModel* 	    producer_model,
				       GdamPlayInfo*        play_info,
			               int           	    new_source_id,
			               GdamRequestBuilder*  builder)
{
	GdamRequest* request;
	GdamTurntable* turntable = GDAM_TURNTABLE(turntable_model);
	GdamProducer* producer = GDAM_PRODUCER(producer_model);
	int resample_id_tmp, gain_id_tmp;

	gdam_debug("gdam_turntable_new_request: starting to play...");

	g_return_if_fail(turntable->producer == producer);

	if (new_source_id == 0) {
		GdamArg arg;
		int source_id = turntable_model->source_id;
		if (source_id)
			append_unsubscribe_events(turntable, builder);

		arg.type = GDAM_ARG_TYPE_INT;
		arg.name = "filtered";
		GDAM_ARG_INT(&arg) = 0;

		request = gdam_request_set_args(source_id,
		                                1,
						&arg,
						NULL,
						NULL,
						NULL);
		gdam_request_builder_append_generic(builder, request);
		gdam_model_new_child_request(turntable_model->parent,
		                                 turntable_model,
						 play_info,
						 0,
						 builder);
		gdam_model_set_source_id(turntable_model, 0);
		return;
	}

	append_unsubscribe_events(turntable, builder);

	request = build_resample_request(turntable, new_source_id);
	resample_id_tmp = gdam_request_builder_append_create(builder, request);
	request = build_gain_request(turntable, resample_id_tmp);
	gain_id_tmp = gdam_request_builder_append_create(builder, request);

	append_subscribe_events(turntable, builder, new_source_id);

	gdam_debug("turntable [%p]: calling `new_child_request' with parent %p",
				turntable_model,
				turntable_model->parent);
	gdam_model_new_child_request(turntable_model->parent,
	                                 turntable_model,
					 play_info,
					 gain_id_tmp,
					 builder);
	gdam_debug("turntable: done calling `new_child_request' with parent");
}

static void gdam_turntable_play(GdamProducer*  producer,
				GdamPlayInfo*  play_info)
{
	GdamTurntableClass* tt_class;
	GdamTurntable* turntable = GDAM_TURNTABLE(producer);
	GdamPlayInfo* a_play_info =   play_info 
				  ? play_info 
				  : g_new0(GdamPlayInfo, 1);
	tt_class = GDAM_TURNTABLE_CLASS(GTK_OBJECT(turntable)->klass);

	if (GDAM_MODEL(turntable)->source_id != 0) {
		turntable_do_sync(turntable);
		return;
	}
	if (turntable->producer == NULL) {
		g_warning("no producer... aborting turntable_play_song");
		return;
	}
		/* ??? */
	a_play_info->position = 
		GDAM_PRODUCER(turntable)->stopped_position * turntable->rate;
	gdam_producer_play(turntable->producer,
			   a_play_info);
}

/* TODO: forward stopped_pos to sub producer. */
static void gdam_turntable_real_set_position(GdamProducer* ttproducer,
					     gdouble	   position)
{
	GdamTurntable* turntable = GDAM_TURNTABLE(ttproducer);
	GdamProducer* sub_prod = turntable->producer;
	float sub_pos;
	ttproducer->stopped_position = position;
	sub_pos = position * turntable->rate;
	gdam_producer_set_position(sub_prod, sub_pos);
	gtk_signal_emit(GTK_OBJECT(turntable),
			turntable_signals[UPDATE_BEATNUM],
			gdam_turntable_get_beat_number(turntable, position));
}

/* Changes the turntable's resampling rate so that the apparent tempo of the
 *   output sound matches the argument.
 */
void turntable_set_tempo(GdamTurntable* turntable, gdouble new_tempo)
{
	float rate;
	GdamChannel* channel = ((GdamModel*)turntable)->channel;
	GdamModel* model = GDAM_MODEL(turntable);
	GdamModel* prod_model = GDAM_MODEL(turntable->producer);
	GdamBeatInfo new_beat;
	float comp_bpm = gdam_beat_info_get_bpm(&model->beat_info,0);
	float prod_bpm = gdam_beat_info_get_bpm(&prod_model->beat_info,0);
	float prod_fb;
	float bpm_factor = turntable->bpm_factor;
	prod_fb = gdam_beat_info_get_first_beat(&prod_model->beat_info,0);
	if (comp_bpm <= 0.0 || turntable->producer == NULL)
		return;
	rate = new_tempo * bpm_factor / prod_bpm;

	gdam_debug("turntable: matching to %f", new_tempo);

	if (turntable->resample_id) {
		GdamArg arg;
		GdamRequest* request;
		arg.name = "rate";
		arg.type = GDAM_ARG_TYPE_DOUBLE;
		GDAM_ARG_DOUBLE(&arg) = rate;
		request = gdam_request_set_args(
				turntable->resample_id,
				1,
				&arg,
				gdam_do_nothing,
				gdam_do_nothing,
				NULL);
		gdam_channel_append_request(channel, request);
	}
	if (!turntable->producer) {
		g_warning("no producing submodel");
		return;
	}

	turntable->rate = rate;

	gdam_beat_info_set_bpm(&new_beat, prod_bpm * rate);
	gdam_beat_info_set_first_beat(&new_beat, prod_fb / rate);

	gdam_model_set_beat_info(model, &new_beat);

	gtk_signal_emit(GTK_OBJECT(turntable),
			turntable_signals[SET_RATE],
			rate);
}
GdamProducer* gdam_turntable_get_producer(GdamTurntable*      turntable)
{
	return turntable->producer;
}

void gdam_turntable_set_producer(GdamTurntable*      turntable,
                                 GdamProducer*       producer)
{
	if (turntable->producer) {
		GtkObject* obj_producer = GTK_OBJECT(turntable->producer);
		//gtk_signal_disconnect(obj_producer,
		                      //turntable->sub_set_source_handle);
		GDAM_MODEL(turntable->producer)->parent = NULL;
		gtk_signal_disconnect(obj_producer,
		                      turntable->sub_set_beat_info_handle);
		gtk_signal_disconnect(obj_producer,
		                      turntable->sub_set_position_handle);
		gtk_signal_disconnect(obj_producer,
		                      turntable->sub_update_length_handle);
		gtk_object_unref(obj_producer);
	}
	turntable->producer = producer;
	if (producer) {
		GtkSignalFunc handle_func;
		GDAM_MODEL(producer)->parent = GDAM_MODEL(turntable);
		gtk_object_ref(GTK_OBJECT(producer));
		gtk_object_sink(GTK_OBJECT(producer));
		handle_func = GTK_SIGNAL_FUNC(handle_producer_set_position);
		turntable->sub_set_position_handle
			= gtk_signal_connect(GTK_OBJECT(producer),
					     "set_position",
					     handle_func,
					     turntable);
		handle_func = GTK_SIGNAL_FUNC(handle_producer_update_length);
		turntable->sub_update_length_handle
			= gtk_signal_connect(GTK_OBJECT(producer),
					     "update_length",
					     handle_func,
					     turntable);
		handle_func = GTK_SIGNAL_FUNC(handle_producer_set_beat_info);
		turntable->sub_set_beat_info_handle
			= gtk_signal_connect(GTK_OBJECT(producer),
					     "set_beat_info",
					     handle_func,
					     turntable);
	}
}


float gdam_turntable_get_beat_number(GdamTurntable* turntable, float _netpos)
{
	GdamModel* model;
	double since_first_beat;
	float netpos = _netpos;
	//double bpm, first_beat;
	//gdam_debug("turntable: getting beat number for time %f.", netpos);
	if (!turntable || !turntable->producer){
		gdam_debug("get_beat_number aborting");
		return 0.0;
	}
	model = GDAM_MODEL(turntable);
	since_first_beat = netpos - 
			gdam_beat_info_get_first_beat(&model->beat_info,0);
	return since_first_beat *
			 gdam_beat_info_get_bpm(&model->beat_info,0) / 60.0;
}

void gdam_turntable_zero(GdamTurntable* turntable)
{
	GdamModel* producer = GDAM_MODEL(turntable->producer);
	if (!producer) {
		g_warning("gdam_turntable_zero: no producer ");
		return;
	}
	if (producer) {
		if (producer->source_id) {
			GdamPlayInfo* play_info = g_new0(GdamPlayInfo,1);
			play_info->position = 0.0;
			gdam_producer_play(turntable->producer, play_info);
			g_free(play_info);
		} else {
			GDAM_PRODUCER(turntable)->stopped_position = 0.0;
		}
	}
	gdam_producer_set_position(GDAM_PRODUCER(turntable), 0.0);
}

/* Seeks to the first beat of the song. */
void gdam_turntable_start(GdamTurntable* turntable)
{
	GdamModel* ttcomp = GDAM_MODEL(turntable);
	gdouble first_beat;
	gdouble producer_first_beat;
	gdam_debug("gdam_turntable_start");
	if (!turntable->producer) {
		g_warning("gdam_turntable_start: no producer: cannot start");
		return;
	}
	first_beat = gdam_beat_info_get_first_beat(&ttcomp->beat_info,0);
	producer_first_beat = first_beat / turntable->rate;
	if (turntable->producer) {
		if (GDAM_MODEL(turntable->producer)->source_id) {
			GdamPlayInfo* play_info = g_new0(GdamPlayInfo,1);
			play_info->position = producer_first_beat;
			gdam_producer_play(turntable->producer, play_info);
			g_free(play_info);
		} else {
			gdam_producer_set_position(turntable->producer, 
							producer_first_beat);
		}
	}
	gdam_producer_set_position(GDAM_PRODUCER(turntable), first_beat);
}

static gboolean gdam_turntable_match_beat_info(GdamModel* model,
					       const GdamBeatInfo* match_info)
{
	GdamTurntable* turntable = GDAM_TURNTABLE(model);
	float bpm = gdam_beat_info_get_bpm(match_info,0);
	if (!turntable || !match_info)
		g_warning("turntable_match_beat_info: NULL argument.");
	turntable_set_tempo(turntable, bpm);
	return TRUE;
}


static void gdam_turntable_real_set_rate(GdamTurntable* turntable, 
					 gdouble newrate)
{
	gdam_debug("gdam turntable real set rate");
	turntable->rate = newrate;
	resample_rate_changed(turntable);
}

void gdam_turntable_set_rate(GdamTurntable* turntable, float newrate)
{
	GtkObject* this_object = GTK_OBJECT(turntable);

	gdam_debug("gdam turntable set rate");

	if (gtk_signal_n_emissions(this_object,
	                           turntable_signals[SET_RATE]))
		return;

	gtk_signal_emit(this_object,
			turntable_signals[SET_RATE],
			newrate);
}

void gdam_turntable_update_beatnum(GdamTurntable* turntable, float newbn)
{
	GtkObject* this_object = GTK_OBJECT(turntable);

	if (gtk_signal_n_emissions(this_object,
	                           turntable_signals[UPDATE_BEATNUM]))
		return;

	gtk_signal_emit(this_object,
			turntable_signals[UPDATE_BEATNUM],
			newbn);
}


void gdam_turntable_update_jump_preview(GdamTurntable* turntable, float newval)
{
	GtkObject* this_object = GTK_OBJECT(turntable);

	if (gtk_signal_n_emissions(this_object,
	                           turntable_signals[UPDATE_JUMP_PREVIEW]))
		return;

	gtk_signal_emit(this_object,
			turntable_signals[UPDATE_JUMP_PREVIEW],
			newval);
}



static void gdam_turntable_add_xml(GdamModel*     model,
                                   xmlNode*       parent,
				   GdamXmlInfo*   xml_info)
{
	GdamTurntable* turntable;
	GdamProducer* producer;
	turntable = GDAM_TURNTABLE(model);

	producer = turntable->producer;
	if (producer != NULL
	 && gdam_xml_info_lookup_by_object(xml_info, producer) == 0) {
		xmlNode* node;
		xmlNode* stack_node;
		GdamModel* producer = GDAM_MODEL(turntable->producer);
		node = gdam_model_to_xml(producer, xml_info);
		stack_node = xmlNewNode(NULL, "producer");
		xmlAddChild(stack_node, node);
		xmlAddChild(parent, stack_node);
	}
}

static void gdam_turntable_parse_xml(GdamModel*     model,
                                     xmlNode*       node,
				     GdamXmlInfo*   xml_info)
{
	GdamModel* submodel;
	GdamTurntable* turntable;
	turntable = GDAM_TURNTABLE(model);
	if (strcmp(node->name, "producer") == 0) {
		if (node->childs == NULL || node->childs->next != NULL) {
			g_warning("tag <filter> must have exactly one subtag");
			return;
		}
		submodel = gdam_model_from_xml(node->childs,
		                               model->channel,
					       xml_info);
		if (! GDAM_IS_PRODUCER(submodel)) {
			GtkType type = GTK_OBJECT(submodel)->klass->type;
			g_warning("the turntable's producer was set to %s, "
				"which does not derive from GdamProducer",
				gtk_type_name(type));
			return;
		}
		gdam_turntable_set_producer(turntable,
		                            GDAM_PRODUCER(submodel));
		return;
	}

	g_warning("node <%s> unhandler in GdamBuilder", node->name);
}

