/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdammididevice.h"
#include <gdam/gdamsearchpath.h>
#include <gdam/gdamplugin.h>
#include "../../common/gdamdebug.h"
#include <math.h>
#include <gtk/gtksignal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>

/* Maximum number of errors we may encounter in a midi file
 * before stopping processing. */
#define MAX_ERRORS	20

static const char* gdam_midi_note_names[] =
{
	"c", "c#",
	"d", "d#",
	"e",
	"f", "f#",
	"g", "g#",
	"a", "a#",
	"b"
};

enum {
	EVENT,
	NEW_TRAP,
	NOTE_ON,
	NOTE_OFF,
#if 0
	PITCH_WHEEL,
	MIDI_PARAM,
#endif
	LAST_SIGNAL
};

/* A single connection to a trap. */
typedef struct _GdamMidiHook GdamMidiHook;

/* A single attachment to a midi trap. */
struct _GdamMidiHook {
	/* The unique id for this hook on this device. */
	guint			hook_id;

	/* The trap we are connected to. */
	GdamMidiTrap*		trap;

	/* The min/max to set the parameters to. */
	gdouble			scale_max, scale_min;

	/* The function to call when we get a corresponding midi event. */
	GdamMidiHookFunc        hook_func;

	/* The function to call when the hook is destroyed. (optional) */
	GDestroyNotify		destroy;
	gpointer		user_data;
	char*			hook_name;

	/* A object whose destruction will disconnect the hook. */
	GtkObject*		alive_object;
	guint			alive_destroy_id;
};

/* Load the device-table from ~/.gdam/midi.devices or wherever...
 *
 * This describes which (in any) patch files to load when a given
 * midi device (specified by /dev/midi* device name) is opened.
 */
static void maybe_initialize_master_patch_table();

static gboolean gdam_midi_device_has_patchfile(GdamMidiDevice* device, 
                                               const char* patchfilename);
static void gdam_midi_device_add_pitchwheel_name(GdamMidiDevice* device,
                                                 int             channel,
						 const char*     name);
static void gdam_midi_device_add_parameter_name (GdamMidiDevice* device,
                                                 int             channel,
						 int             param,
						 const char*     name);
static void gdam_midi_device_add_sysex_name     (GdamMidiDevice* device,
						 int             make_id,
						 int             model_id,
                                                 int             length,
						 const char*     data,
						 const char*     name);
static gboolean process_midi_device_input(GIOChannel*   io_channel,
                                          GIOCondition  condition,
					  gpointer      user_data);

static int gdam_midi_device_open_fd(GdamMidiDevice* device)
{
	return open(device->filename, O_RDONLY);
}

static guint memhash(const guint8* msg, int length)
{
	guint hash = 5003;
	const guint8* end = msg + length;
	while (msg < end) {
		hash += (hash<<5) + (guint) (*msg);
		msg++;
	}
	return hash;
}

guint gdam_midi_trap_hash(gconstpointer ptr)
{
	const GdamMidiTrap* trap = ptr;
	if (trap->sysex_message != NULL)
		return memhash(trap->sysex_message, trap->sysex_length);
	return ((guint)trap->type << 24)
	     ^ ((guint)trap->channel << 16)
	     ^ ((guint)trap->which_param);
}
gboolean gdam_midi_trap_equal(gconstpointer ptr_1, gconstpointer ptr_2)
{
	const GdamMidiTrap* trap_1 = ptr_1;
	const GdamMidiTrap* trap_2 = ptr_2;
	if (trap_1->type != trap_2->type)
		return FALSE;
	if (trap_1->sysex_message != NULL && trap_2->sysex_message != NULL)
		return trap_1->sysex_length == trap_2->sysex_length
		    && memcmp(trap_1->sysex_message,
		              trap_2->sysex_message,
		              trap_1->sysex_length) == 0;

	return    trap_1->channel == trap_2->channel
	  &&  trap_1->which_param == trap_2->which_param;
}

char* gdam_midi_trap_get_string(GdamMidiTrap* trap)
{
	const char* format = NULL;
	if (trap->type == GDAM_MIDI_PARAMETER) {
		format = "%s, channel %d, parameter %d";
	} else if (trap->type == GDAM_MIDI_PITCH_WHEEL) {
		format = "%s, channel %d, pitchwheel";
	} else if (trap->type == GDAM_MIDI_SYSEX) {
		GString* string = g_string_new(NULL);
		char* rv;
		int i;
		g_string_sprintf(string, "%s,", trap->device->filename);
		for (i = 0; i < trap->sysex_length; i++) {
			int d = trap->sysex_message[i];
			g_string_sprintfa(string, " 0x%02x", d);
		}
		rv = string->str;
		g_string_free(string, FALSE);
		return rv;
	} else {
		return g_strdup("unknown midi trap type");
	}
	return g_strdup_printf(format, trap->device->filename,
	                               trap->channel,
				       trap->which_param);
}

static void get_model_data(int       make_id,
		           int       model_id,
			   char**    make_name,
			   char**    model_name);

static guint midi_signals[LAST_SIGNAL] = { 0 };

static GdamMidiPitch* gdam_midi_pitch_new (int		channel,
					   int		pitch)
{
	GdamMidiPitch* mpitch = g_new(GdamMidiPitch, 1);
	mpitch->type = GDAM_MIDI_PITCH_WHEEL;
	mpitch->channel = channel;
	mpitch->pitch = pitch;

	/* XXX:  the `trap' contains better data about the range. */
	mpitch->norm_value = (gdouble)pitch / (1<<14);
	return mpitch;
}

static GdamMidiParam* gdam_midi_param_new (int		channel,
					   int		parameter,
					   int		value)
{
	GdamMidiParam* param = g_new(GdamMidiParam, 1);
	param->type = GDAM_MIDI_PARAMETER;
	param->channel = channel;
	param->parameter = parameter;
	param->value = value;
	/* XXX:  the `trap' contains better data about the range. */
	param->norm_value = (gdouble)value / (1<<7);
	return param;
}

static void            gdam_midi_pitch_destroy(GdamMidiPitch*    to_destroy)
{
	g_free(to_destroy);
}

static void            gdam_midi_param_destroy(GdamMidiParam*    to_destroy)
{
	g_free(to_destroy);
}

GdamMidiNote*	gdam_midi_note_new    (int              instrument_id,
                                       int		midi_id,
                                       int              velocity)
{
	GdamMidiNote* note = g_new(GdamMidiNote, 1);
	note->type = GDAM_MIDI_NOTE_ON;
	note->instrument = instrument_id;
	note->midi_index = midi_id;
	note->note_name = gdam_midi_note_names[midi_id % 12];
	note->octave = (midi_id - 12) / 12;
	note->hertz = 440.0 * pow(2.0, (gdouble)(midi_id - 57.0) / 12.0);
	note->velocity = velocity;
	return note;
}

void            gdam_midi_note_destroy(GdamMidiNote*    note_to_destroy)
{
	g_free(note_to_destroy);
}


static void gdam_midi_device_init(GdamMidiDevice* midi_device)
{
}
static void gdam_midi_device_class_init(GdamMidiDeviceClass* midi_device_class)
{
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(midi_device_class);
	GtkType type = object_class->type;

	midi_device_class->parse_raw = process_midi_device_input;
	midi_device_class->open_fd = gdam_midi_device_open_fd;

	midi_signals[EVENT] = gtk_signal_new("event",
	                                     GTK_RUN_LAST,
					     type,
					     GTK_SIGNAL_OFFSET(
					       	 GdamMidiDeviceClass,
						 event
					     ),
					     gtk_marshal_NONE__POINTER,
					     GTK_TYPE_NONE,
					     1,
					     GTK_TYPE_POINTER);
	midi_signals[NEW_TRAP] = gtk_signal_new("new-trap",
	                                     GTK_RUN_LAST,
					     type,
					     GTK_SIGNAL_OFFSET(
					       	 GdamMidiDeviceClass,
						 new_trap
					     ),
					     gtk_marshal_NONE__POINTER_POINTER,
					     GTK_TYPE_NONE,
					     2,
					     GTK_TYPE_POINTER,
					     GTK_TYPE_POINTER);
		
		
	midi_signals[NOTE_ON] = gtk_signal_new("note_on",
	                                       GTK_RUN_LAST,
					       type,
					       GTK_SIGNAL_OFFSET(
					       	 GdamMidiDeviceClass,
						 note_on
					       ),
					       gtk_marshal_NONE__POINTER,
					       GTK_TYPE_NONE,
					       1,
					       GTK_TYPE_POINTER);
	midi_signals[NOTE_OFF] = gtk_signal_new("note_off",
	                                       GTK_RUN_LAST,
					       type,
					       GTK_SIGNAL_OFFSET(
					       	 GdamMidiDeviceClass,
						 note_off
					       ),
					       gtk_marshal_NONE__POINTER,
					       GTK_TYPE_NONE,
					       1,
					       GTK_TYPE_POINTER);
	gtk_object_class_add_signals(object_class, midi_signals, LAST_SIGNAL);
}

GtkType gdam_midi_device_get_type()
{
	static GtkType midi_device_type = 0;
	if (!midi_device_type) {
		static const GtkTypeInfo midi_device_info = {
			"GdamMidiDevice",
			sizeof(GdamMidiDevice),
			sizeof(GdamMidiDeviceClass),
			(GtkClassInitFunc) gdam_midi_device_class_init,
			(GtkObjectInitFunc) gdam_midi_device_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GTK_TYPE_OBJECT;
		midi_device_type = gtk_type_unique(parent,
		                                 &midi_device_info);
	}
	return midi_device_type;
}

GdamMidiDevice* gdam_midi_device_new ()
{
	return GDAM_MIDI_DEVICE( gtk_type_new(GDAM_TYPE_MIDI_DEVICE) );
}

/*
 *  ____                            ____  _             _
 * |  _ \ __ _ _ __ ___  ___ _ __  |  _ \| |_   _  __ _(_)_ __  ___
 * | |_) / _` | '__/ __|/ _ \ '__| | |_) | | | | |/ _` | | '_ \/ __|
 * |  __/ (_| | |  \__ \  __/ |    |  __/| | |_| | (_| | | | | \__ \
 * |_|   \__,_|_|  |___/\___|_|    |_|   |_|\__,_|\__, |_|_| |_|___/
 *                                                |___/
 */
static GHashTable* plugins_by_device = NULL;
static const char* get_type_for_device(const char* device_name)
{
	if (plugins_by_device == NULL)
		return NULL;
	return (const char*)g_hash_table_lookup(plugins_by_device,
	                                        device_name);
}
static void assoc_module_with_device(const char* device_name,
                                     const char* module_name)
{
	char* old;

	if (plugins_by_device == NULL)
		plugins_by_device = g_hash_table_new(g_str_hash, g_str_equal);
	old = g_hash_table_lookup(plugins_by_device, device_name);
	if (old != NULL) {
		g_hash_table_insert(plugins_by_device,
		                    (char*) device_name,
		                    g_strdup(module_name));
		g_free(old);
	} else {
		g_hash_table_insert(plugins_by_device,
		                    g_strdup(device_name),
		                    g_strdup(module_name));
	}
}

static GdamMidiTrap*   try_trap    (GdamMidiDevice*       device,
				    GdamMidiEventType     type,
				    int                   channel,
				    int                   which_param)
{
	GdamMidiTrap hash_part;
	hash_part.type = type;
	hash_part.channel = channel;
	hash_part.which_param = which_param;
	hash_part.sysex_message = NULL;
	hash_part.sysex_length = 0;
	if (device->trap_table == NULL)
		return NULL;
	return g_hash_table_lookup(device->trap_table, &hash_part);
}

static void force_trap_table_to_exist(GdamMidiDevice* device)
{
	if (device->trap_table == NULL) {
		device->trap_table = g_hash_table_new(gdam_midi_trap_hash,
		                                      gdam_midi_trap_equal);
		device->hook_id_table = g_hash_table_new(NULL, NULL);
	}
}

static GdamMidiTrap*   force_trap  (GdamMidiDevice*       device,
				    GdamMidiEventType     type,
				    int                   channel,
				    int                   which_param)
{
	GdamMidiTrap hash_part;
	GdamMidiTrap* rv;
	hash_part.type = type;
	hash_part.channel = channel;
	hash_part.which_param = which_param;
	hash_part.sysex_message = NULL;
	hash_part.sysex_length = 0;
	force_trap_table_to_exist(device);
	rv = g_hash_table_lookup(device->trap_table, &hash_part);
	if (rv == NULL) {
		rv = g_new0(GdamMidiTrap, 1);
		rv->type = type;
		rv->channel = channel;
		rv->which_param = which_param;
		rv->min_value = 0;
		rv->device = device;
		if (type == GDAM_MIDI_PITCH_WHEEL)
			rv->max_value = (1<<14)-1;
		else
			rv->max_value = 127;
		g_hash_table_insert(device->trap_table, rv, rv);
	}
	return rv;
}

static GdamMidiTrap*   try_sysex_trap  (GdamMidiDevice*       device,
					const char*           make_name,
					const char*           model_name,
				        int                   length,
				        const guint8*         data)
{
	GdamMidiTrap hash_part;
	GdamMidiTrap* rv;
	if (device->trap_table == NULL)
		return NULL;
	hash_part.type = GDAM_MIDI_SYSEX;
	hash_part.channel = 0;
	hash_part.which_param = 0;
	hash_part.sysex_message = (guint8*) data;
	hash_part.sysex_length = length;
	hash_part.make_name = make_name;
	hash_part.model_name = model_name;
	rv = g_hash_table_lookup(device->trap_table, &hash_part);
	return rv;
}

static GdamMidiTrap*   force_sysex_trap  (GdamMidiDevice*       device,
					  const char*           make_name,
					  const char*           model_name,
				          int                   length,
				          const guint8*         data)
{
	GdamMidiTrap hash_part;
	GdamMidiTrap* rv;
	hash_part.type = GDAM_MIDI_SYSEX;
	hash_part.channel = 0;
	hash_part.which_param = 0;
	hash_part.sysex_message = (guint8*) data;
	hash_part.sysex_length = length;
	hash_part.make_name = make_name;
	hash_part.model_name = model_name;
	force_trap_table_to_exist(device);
	rv = g_hash_table_lookup(device->trap_table, &hash_part);
	if (rv == NULL) {
		rv = g_new0(GdamMidiTrap, 1);
		rv->type = GDAM_MIDI_SYSEX;
		rv->channel = 0;
		rv->which_param = 0;
		rv->min_value = 0;
		rv->device = device;
		rv->sysex_message = g_memdup(data, length);
		rv->sysex_length = length;
		rv->make_name = make_name;
		rv->model_name = model_name;
		g_hash_table_insert(device->trap_table, rv, rv);
	}
	return rv;
}


GdamMidiTrap* gdam_midi_device_get_trap(GdamMidiDevice* device,
                                        GdamMidiEvent*  event,
					gboolean        create_trap)
{
	int channel = 0, which_param = 0;
	switch (event->type) {
		case GDAM_MIDI_PITCH_WHEEL:
			channel = event->pitch_event.channel;
			break;
		case GDAM_MIDI_PARAMETER:
			channel = event->param_event.channel;
			which_param = event->param_event.parameter;
			break;
		case GDAM_MIDI_SYSEX: {
			GdamMidiSystemExclusive* sysex = &event->sysex_event;
			if (create_trap)
				return force_sysex_trap(device,
							sysex->make_name,
							sysex->model_name,
				                        sysex->message_length,
							sysex->message);
			else
				return try_sysex_trap(device,
						      sysex->make_name,
						      sysex->model_name,
				                      sysex->message_length,
					              sysex->message);
		}
		default: {
			return NULL;
		}
	}
	if (create_trap)
		return force_trap(device, event->type, channel, which_param);
	return try_trap(device, event->type, channel, which_param);
}

static void marshal_trap(GdamMidiDevice*        device,
                         GdamMidiTrap*          trap,
			 GdamMidiEvent*         event)
{
	GSList* at;

	if (trap == NULL || trap->hooks == NULL)
		return;

	for (at = trap->hooks; at != NULL; at = at->next) {
		GdamMidiHook* hook = at->data;
		(*hook->hook_func)(device, event, hook->user_data);
	}
}
static void emit_event(GdamMidiDevice*  device,
                       GdamMidiEvent*   event)
{
	gtk_signal_emit(GTK_OBJECT(device), 
	                midi_signals[EVENT],
			event);
}

static void emit_new_trap(GdamMidiDevice* device,
                          GdamMidiTrap*   trap,
			  GdamMidiEvent*  event)
{
	gtk_signal_emit(GTK_OBJECT(device),
	                midi_signals[NEW_TRAP],
			trap,
			event);
}

void gdam_midi_device_emit_program      (GdamMidiDevice* device,
                                         int             channel,
				         int             program)
{
	gdam_debug("MIDI: Program: channel %d set to program %d [ignored]",
				channel, program);
}

void gdam_midi_device_emit_channel_pressure(GdamMidiDevice* device,
                                            int             channel,
					    int             pressure)
{
	gdam_debug("MIDI: Channel Pressure: "
	          "channel %d set to program %d [ignored]",
				channel, pressure);
}

void gdam_midi_device_emit_note_off      (GdamMidiDevice* device,
					  int             instrument,
                                          int             pitch,
					  int             volume)
{
	GdamMidiNote* note;
	note = gdam_midi_note_new(instrument, pitch, volume);
	gdam_debug("note_off: instrument %d, %s octave %d, volume %d",
		instrument,
		note->note_name,
		note->octave,
		note->velocity);
	gtk_signal_emit(GTK_OBJECT(device),
	                midi_signals[NOTE_OFF],
			note);
	emit_event(device, (GdamMidiEvent*) note);
	gdam_midi_note_destroy(note);
}
void gdam_midi_device_emit_note_on       (GdamMidiDevice* device,
					  int             instrument,
                                          int             pitch,
					  int             volume)
{
	GdamMidiNote* note;
	if (volume == 0) {
		gdam_midi_device_emit_note_off(device,
		                               instrument, 
		                               pitch,
					       volume);
		return;
	}
	note = gdam_midi_note_new(instrument, pitch, volume);
	gdam_debug("note_on: instrument %d, %s octave %d, volume %d",
		instrument,
		note->note_name,
		note->octave,
		note->velocity);
	gtk_signal_emit(GTK_OBJECT(device),
	                midi_signals[NOTE_ON],
			note);
	emit_event(device, (GdamMidiEvent*) note);
	gdam_midi_note_destroy(note);
}
void gdam_midi_device_emit_key_pressure       (GdamMidiDevice* device,
					       int             instrument,
                                               int             pitch,
					       int             volume)
{
#if 0
	GdamMidiNote* note;
	note = gdam_midi_note_new(instrument, pitch, volume);
	gtk_signal_emit(GTK_OBJECT(device),
	                midi_signals[KEY_PRESSURE],
			note);
	gdam_midi_note_destroy(note);
#endif
g_warning("MIDI: Key Pressure: ignoring this, but really shouldn't");
}
	
void gdam_midi_device_emit_parameter       (GdamMidiDevice* device,
                                            int             channel,
					    int             which_param,
					    int             value)
{
	GdamMidiParam* param;
	GdamMidiTrap* trap;

	param = gdam_midi_param_new(channel, which_param, value);

	/*gtk_signal_emit(GTK_OBJECT(device),
	                midi_signals[MIDI_PARAM],
			param);*/

	trap = try_trap(device, GDAM_MIDI_PARAMETER, channel, which_param);
	if (trap != NULL)
		marshal_trap(device, trap, (GdamMidiEvent*) param);
	else {
		trap = force_trap(device,
		                  GDAM_MIDI_PARAMETER,
				  channel,
				  which_param);
		emit_new_trap(device, trap, (GdamMidiEvent*) param);
	}

	emit_event(device, (GdamMidiEvent*) param);
	gdam_midi_param_destroy(param);

	gdam_debug("MIDI: Parameter: channel %d, param %d set to %d",
			channel, which_param, value);
}

void gdam_midi_device_emit_pitch_wheel(GdamMidiDevice* device,
                                       int             channel,
				       int             pitch)
{
	GdamMidiTrap* trap;
	GdamMidiPitch* mpitch;

	mpitch = gdam_midi_pitch_new(channel, pitch);

	trap = try_trap(device, GDAM_MIDI_PITCH_WHEEL, channel, 0);
	if (trap != NULL)
		marshal_trap(device, trap, (GdamMidiEvent*) mpitch);
	else {
		trap = force_trap(device,
		                  GDAM_MIDI_PITCH_WHEEL,
				  channel,
				  0);
		emit_new_trap(device, trap, (GdamMidiEvent*) mpitch);
	}

	emit_event(device, (GdamMidiEvent*) mpitch);

	gdam_debug("MIDI: PitchWheel: channel %d, set to %d",
			channel, pitch);
	gdam_midi_pitch_destroy(mpitch);

}
void gdam_midi_device_emit_sysex               (GdamMidiDevice* device,
				                const char*     maker,
				                const char*     model,
                                                int             length,
				                const guint8*   bytes)
{
	GdamMidiTrap* trap;
	GdamMidiSystemExclusive sysex;

	if (length < 0)
		return;

	sysex.type = GDAM_MIDI_SYSEX;
	sysex.make_name = maker;
	sysex.model_name = model;
	sysex.message_length = length;
	sysex.message = bytes;

	/* XXX: should contain the maker/model !!! */
	trap = try_sysex_trap(device, maker, model, length, bytes);
	if (trap != NULL)
		marshal_trap(device, trap, (GdamMidiEvent*) &sysex);
	else {
		trap = force_sysex_trap
		                 (device, maker, model, length, bytes);
		emit_new_trap(device, trap, (GdamMidiEvent*) &sysex);
	}

	emit_event(device, (GdamMidiEvent*) &sysex);
}

static void add_to_triplet(GdamMidiDevice* device, int c)
{
	if (device->triplet_length >= 3) {
		g_warning("midi protocol error -- command too long");
		return;
	}
	if (device->triplet_length == 0
	 && device->triplet[0] != '\0'
	 && c < 0x80) {
	 	device->triplet_length++;
	}

	device->triplet[device->triplet_length++] = (unsigned char) c;
}

static void maybe_process_triplet(GdamMidiDevice* device)
{
	int hi, lo, arg1, arg2;
	if (device->triplet_length == 0)
		return;
	if (device->triplet_length == 1) {
	 	if (device->triplet[0] < 0x80) {
			g_warning("protocol error: "
			          "command didn't begin with >0x80");
	 		device->triplet_length = 0;
		}
		else if (device->triplet[0] >= 0xF0 
					 && device->triplet[0] < 0xFF) {
			g_warning("not parsing: %d", device->triplet[0]);
	 		device->triplet_length = 0;
		}
		return;
	}

	hi = (device->triplet[0] / 16);
	lo = (device->triplet[0] % 16);

	arg1 = device->triplet[1];
	if (device->triplet_length == 2) {
		switch (hi) {
			case 12:
				gdam_midi_device_emit_program(device, lo, arg1);
				device->triplet_length = 0;
				break;
			case 13:
				gdam_midi_device_emit_channel_pressure
							(device, lo, arg1);
				device->triplet_length = 0;
				break;
			default:
		}
		return;
	}

	arg2 = device->triplet[2];
	switch (hi) {
		case 8:
			gdam_midi_device_emit_note_off(device, lo, arg1, arg2);
			device->triplet_length = 0;
			break;
		case 9:
			gdam_midi_device_emit_note_on(device, lo, arg1, arg2);
			device->triplet_length = 0;
			break;
		case 10:
			gdam_midi_device_emit_key_pressure(device, lo, arg1, arg2);
			device->triplet_length = 0;
			break;
		case 11:
			gdam_midi_device_emit_parameter(device, lo, arg1, arg2);
			device->triplet_length = 0;
			break;
		case 14: {
			int pitch;
			pitch = (arg1 & 0x7f)
			      | ((arg2 & 0x7f) << 7);
			gdam_midi_device_emit_pitch_wheel(device, lo, pitch);
			device->triplet_length = 0;
			break;
		}
	}
	if (device->triplet_length > 0) {
		g_warning("error parsing midi commands");
		device->triplet_length = 0;
		return;
	}
	return;
}
static void append_to_system_exclusive(GdamMidiDevice*  device,
		                       int              c)
{
	if (device->system_exclusive == NULL) {
		char str[2];
		str[0] = c;
		str[1] = '\0';
		device->system_exclusive = g_string_new(str);
		return;
	}
	device->system_exclusive
			= g_string_append_c( device->system_exclusive,
					     (gchar) c);
}

static void process_system_exclusive  (GdamMidiDevice*  device)
{
	char* str;
	char* make_name, *model_name;
	int left;
	g_return_if_fail(device->system_exclusive != NULL);
	str = device->system_exclusive->str;
	g_return_if_fail(str != NULL);
	left = device->system_exclusive->len;

	if (left < 6) {
		g_message("system_exclusive too short, %d < 6", left);
		g_string_free(device->system_exclusive, TRUE);
		device->system_exclusive = NULL;
		return;
	}

	get_model_data(str[1], 
			  ((int)(unsigned char)str[3] << 8) 
			+ ((int)(unsigned char)str[4]),
			&make_name,
			&model_name);

	gdam_midi_device_emit_sysex(device, 
				    make_name,
				    model_name,
				    left - 6, 	/* 5 + checksum */
				    str + 5);
	g_string_free(device->system_exclusive, TRUE);
	device->system_exclusive = NULL;
}

static gboolean process_midi_device_input(GIOChannel*   io_channel,
                                          GIOCondition  condition,
					  gpointer      user_data)
{
	GdamMidiDevice* device = GDAM_MIDI_DEVICE(user_data);
	unsigned char midi_buf[1024];
	int num_read;
	int index;
	GIOError io_error;
	io_error = g_io_channel_read(io_channel,
	                             midi_buf,
			             sizeof(midi_buf),
			             &num_read);
	if (io_error == G_IO_ERROR_AGAIN)
		return TRUE;
	if (io_error != G_IO_ERROR_NONE) {
		g_warning("error reading from midi device %s",
		          device->filename);
		return FALSE;
	}
	if (num_read <= 0) {
		g_warning("MIDI device closed...");
		return FALSE;
	}

	for (index = 0; index < num_read; index++) {
		int c = midi_buf[index];
		/* This character is used by the midi device 
		 * to keep time. */
		/* TODO: obey these signals... */
		if (c >= 0xf8)
			continue;

		if (c == 0xf0 || device->system_exclusive != NULL) {
			int old_c = c;

			/* When we get 0xf0 in the middle,
			 * we want to transmit two sysex messages,
			 * as though they had not been joined:
			 *
			 *       0xf0  a b c  0xf0  d e f  0xf7
			 *
			 * should become 2 sysex events:
			 *       0xf0  a b c  0xf7
			 *       0xf0  d e f  0xf7
			 */
			if (c == 0xf0 && device->system_exclusive != NULL)
				c = 0xf7;
			append_to_system_exclusive(device, c);
			if (c == 0xf7) {
				process_system_exclusive(device);
			}
			if (old_c == 0xf0)
				device->system_exclusive = g_string_new("\360");
			continue;
		}

		add_to_triplet(device, c);
		maybe_process_triplet(device);
	}
	return TRUE;
}

static gboolean is_in_list_of_strings(GSList* list, const char* fname)
{
	for (   ; list != NULL; list = list->next)
		if (strcmp(fname, (char*)(list->data)) == 0)
			return TRUE;
	return FALSE;
}

static gboolean gdam_midi_device_has_patchfile(GdamMidiDevice* device, 
                                               const char* patchfilename)
{
	return is_in_list_of_strings(device->patch_list, patchfilename);
}

typedef struct _PendingPatches PendingPatches;
struct _PendingPatches {
	char* midi_device_name;
	GSList* patch_filenames;
};

static GSList* pending_patches = NULL;
/*
 *                     add_pending_patch
 *
 * Add a patch file that should be parsed when the device is
 * opened.  We won't store duplicates.
 */
static void add_pending_patch(const char* device_name,
                              const char* mapfile)
{
	GSList* patches;
	PendingPatches* pending;
	for (patches = pending_patches;
	     patches != NULL;
	     patches = patches->next) {
	     	PendingPatches* p = patches->data;
		if (strcmp(device_name, p->midi_device_name) == 0) {
			if (!is_in_list_of_strings(p->patch_filenames, mapfile))
				p->patch_filenames = g_slist_prepend
					(p->patch_filenames, g_strdup(mapfile));
			return;
		}
	}

	pending = g_new0(PendingPatches, 1);
	pending->midi_device_name = g_strdup(device_name);
	pending->patch_filenames = g_slist_prepend(pending->patch_filenames,
	                                           g_strdup(mapfile));
	pending_patches = g_slist_prepend(pending_patches, pending);
}

static void load_map_file(const char* fname, GdamMidiDevice* device)
{
	gdam_midi_device_load_map(device, fname);
}

static void add_pending_patches(GdamMidiDevice* device)
{
	GSList* pp = pending_patches;
	while (pp != NULL) {
		PendingPatches* pending = pp->data;
		if (strcmp(pending->midi_device_name, device->filename) == 0) {
			g_slist_foreach(pending->patch_filenames,
			                (GFunc) load_map_file,
					device);
			pending_patches = g_slist_remove(pending_patches,
			                                 pending);
			return;
		}
		pp = pp->next;
	}
}

gboolean        gdam_midi_device_open(GdamMidiDevice* midi_device,
                                      const char*     device_name)
{
	char* dev_name = g_strdup(device_name);
	int fd;
	GdamMidiDeviceClass* klass;

	klass = GDAM_MIDI_DEVICE_CLASS(GTK_OBJECT(midi_device)->klass);
	if (midi_device->io_channel != NULL)
		gdam_midi_device_close(midi_device);
	
	if (midi_device->filename != NULL)
		g_free(midi_device->filename);
	midi_device->filename = dev_name;

	fd = (*klass->open_fd)(midi_device);
	if (fd < 0) {
		g_warning("couldn't open %s: %s",
			midi_device->filename,
			strerror(errno));
		return FALSE;
	}
	midi_device->io_channel = g_io_channel_unix_new(fd);
	g_io_add_watch(midi_device->io_channel,
	               G_IO_IN,
		       klass->parse_raw,
		       midi_device);
	return TRUE;
}

void            gdam_midi_device_close(GdamMidiDevice* midi_device)
{
	if (midi_device->io_channel) {
		/* XXX: verify that the watch we added is destroyed
		        automatically. */
		g_io_channel_unref(midi_device->io_channel);
		midi_device->io_channel = NULL;
	}
}

/* Table of all opened midi devices. */
static GHashTable*  midi_devices = NULL;
static GHashTable*  device_to_name = NULL;


static void remove_from_hash(GdamMidiDevice* device)
{
	char* name = g_hash_table_lookup(device_to_name, device);
	g_return_if_fail(name != NULL);
	g_hash_table_remove(device_to_name, device);
	g_hash_table_remove(midi_devices, name);
	g_free(name);
}

GdamMidiDevice* gdam_midi_device_get_unique(const char* device_name)
{
	GdamMidiDevice* midi_device;
	char* dev_name_copy;
	const char* type_name;
	maybe_initialize_master_patch_table();
	if (midi_devices == NULL) {
		midi_devices = g_hash_table_new(g_str_hash, g_str_equal);
		device_to_name = g_hash_table_new(NULL, NULL);
	}
	midi_device = g_hash_table_lookup(midi_devices, device_name);
	if (midi_device != NULL) {
		g_message("device %s already open", device_name);
		return midi_device;
	}
	type_name = get_type_for_device(device_name);
	if (type_name == NULL)
		midi_device = gdam_midi_device_new ();
	else {
		GtkType type = gdam_plugin_get_type_by_name("component",
		                                            type_name);
		if (type == 0) {
			g_warning("couldn't open plugin of type %s", type_name);
			return NULL;
		}
		midi_device = GDAM_MIDI_DEVICE(gtk_type_new(type));
	}
	if (! gdam_midi_device_open(midi_device, device_name) ) {
		g_warning("midi open failed...");
		return NULL;
	}
	dev_name_copy = g_strdup(device_name);
	g_hash_table_insert(midi_devices, dev_name_copy, midi_device);
	g_hash_table_insert(device_to_name, midi_device, dev_name_copy);
	gtk_signal_connect(GTK_OBJECT(midi_device),
	                   "destroy",
			   GTK_SIGNAL_FUNC(remove_from_hash),
			   NULL);
	add_pending_patches(midi_device);
	return midi_device;
}

struct _ModelInfo {
	int model_id;
	char* model_name;
};

typedef struct _MakeInfo MakeInfo;
typedef struct _ModelInfo ModelInfo;
struct _MakeInfo {
	int make_id;
	char* make_name;
	GHashTable* models_by_id;
};

GHashTable* all_makes = NULL;

void gdam_midi_device_record_make_name(int id, const char* name)
{
	MakeInfo* info;
	if (all_makes == NULL)
		all_makes = g_hash_table_new(NULL, NULL);
	info = g_hash_table_lookup(all_makes, GINT_TO_POINTER(id));
	if (info != NULL)
		return;
	info = g_new(MakeInfo, 1);
	info->make_id = id;
	info->make_name = g_strdup(name);
	info->models_by_id = NULL;
	g_hash_table_insert(all_makes, GINT_TO_POINTER(id), info);
}
void gdam_midi_device_record_model_name(int make_id, int model_id,
                                        const char* name)
{
	MakeInfo* make_info;
	ModelInfo* model_info;
	g_return_if_fail(all_makes != NULL);
	make_info = g_hash_table_lookup(all_makes, GINT_TO_POINTER(make_id));
	if (make_info == NULL) {
		g_warning("no make of type %d", make_id);
		return;
	}
	model_info = g_new(ModelInfo, 1);
	model_info->model_id = model_id;
	model_info->model_name = g_strdup(name);
	if (make_info->models_by_id == NULL)
		make_info->models_by_id = g_hash_table_new(NULL, NULL);
	g_hash_table_insert(make_info->models_by_id,
	                    GINT_TO_POINTER(model_id),
			    model_info);
	
}

static void get_model_data(int make_id,
		           int model_id,
			   char** make_name,
			   char** model_name)
{
	MakeInfo* make;
	ModelInfo* model;

	if (all_makes == NULL)
		goto unknown;

	make = g_hash_table_lookup(all_makes, GINT_TO_POINTER(make_id));
	if (make == NULL)
		goto unknown;

	if (make->models_by_id != NULL)
		model = g_hash_table_lookup(make->models_by_id, 
		                            GINT_TO_POINTER(model_id));
	else
		model = NULL;
	*make_name = make->make_name;
	*model_name = model ? model->model_name : NULL;
	return;
unknown:

	*make_name = "unknown make";
	*model_name = "unknown model";
	return;
}


static void alive_object_destroyed(GdamMidiHook* hook)
{
	GdamMidiTrap* trap = hook->trap;
	hook->alive_object = NULL;
	gdam_midi_device_disconnect(trap->device, hook->hook_id);
}

static guint    gdam_midi_device_connect_trap
                                     (GdamMidiDevice*    midi_dev,
				      GdamMidiTrap*      trap,
				      const char*        hook_name,
				      GdamMidiHookFunc   hook_func,
				      GDestroyNotify     destroy,
				      gpointer           user_data,
				      GtkObject*         alive_object)
{
	GdamMidiHook* hook;
	guint id;
	hook = g_new0(GdamMidiHook, 1);

	id = hook->hook_id = ++(midi_dev->last_hook_used);
	hook->scale_min = 0.0;
	hook->scale_max = 1.0;
	hook->hook_func = hook_func;
	hook->user_data = user_data;
	hook->destroy   = destroy;
	hook->trap      = trap;
	hook->hook_name = g_strdup(hook_name);
	hook->alive_object = alive_object;

	if (hook->alive_object != NULL) {
		hook->alive_destroy_id = gtk_signal_connect_object
				(alive_object,
				 "destroy",
				 GTK_SIGNAL_FUNC(alive_object_destroyed),
				 (GtkObject*) hook);
	} else {
		hook->alive_destroy_id = 0;
	}

	trap->hooks = g_slist_prepend(trap->hooks, hook);
	g_hash_table_insert(midi_dev->hook_id_table,
	                    GUINT_TO_POINTER(id),
			    hook);
	return id;
}

guint           gdam_midi_device_connect
                                     (GdamMidiDevice*    midi_dev,
				      GdamMidiEventType  type,
				      int                channel,
				      int                which_param,
				      const char*        hook_name,
				      GdamMidiHookFunc   hook_func,
				      GDestroyNotify     destroy,
				      gpointer           user_data,
				      GtkObject*         alive_object)
{
	GdamMidiTrap* trap;
	trap = force_trap(midi_dev, type, channel, which_param);
	g_return_val_if_fail(trap != NULL, 0);

	return gdam_midi_device_connect_trap(midi_dev, trap,
	                                     hook_name,
					     hook_func,
					     destroy,
					     user_data,
					     alive_object);
}

guint           gdam_midi_device_connect_sysex
                                     (GdamMidiDevice*    midi_dev,
				      int                length,
				      const guint8*      data,
				      const char*        hook_name,
				      GdamMidiHookFunc   hook_func,
				      GDestroyNotify     destroy,
				      gpointer           user_data,
				      GtkObject*         alive_object)
{
	GdamMidiTrap* trap;
	const char* make_name = NULL;
	const char* model_name = NULL;
	trap = force_sysex_trap(midi_dev, make_name, model_name, length, data);
	g_return_val_if_fail(trap != NULL, 0);

	return gdam_midi_device_connect_trap(midi_dev,
	                                     trap,
	                                     hook_name,
					     hook_func,
					     destroy,
					     user_data,
					     alive_object);
}

static void gdam_midi_device_add_pitchwheel_name(GdamMidiDevice* device,
                                                 int             channel,
						 const char*     name)
{
	GdamMidiTrap* trap;
	trap = force_trap(device, GDAM_MIDI_PITCH_WHEEL, channel, 0);
	if (trap->trap_name != NULL)
		g_free(trap->trap_name);
	trap->trap_name = g_strdup(name);
}

static void gdam_midi_device_add_parameter_name (GdamMidiDevice* device,
                                                 int             channel,
						 int             param,
						 const char*     name)
{
	GdamMidiTrap* trap;
	trap = force_trap(device, GDAM_MIDI_PARAMETER, channel, param);
	if (trap->trap_name != NULL)
		g_free(trap->trap_name);
	trap->trap_name = g_strdup(name);
}

static void gdam_midi_device_add_sysex_name     (GdamMidiDevice* device,
						 int             make_id,
						 int             model_id,
                                                 int             length,
						 const char*     data,
						 const char*     name)
{
	GdamMidiTrap* trap;
	const char* make_name;
	const char* model_name;
	
	get_model_data(make_id, model_id,
	               (char**) (&make_name), (char**) (&model_name));
	trap = force_sysex_trap(device, make_name, model_name, length, data);
	if (trap->trap_name != NULL)
		g_free(trap->trap_name);
	trap->trap_name = g_strdup(name);
}

static const char* parse_pitch_wheel(GdamMidiDevice* device,
                                     const char* args)
{
	int channel;
	const char* endp;
	const char* description;

	/* Casts are for const portability... */
	channel = strtol( (char*) args, (char**) &endp, 10 );

	if (endp == args)
		return "error parsing pitchwheel channel";
	
	/* Find the description... */
	description = endp;
	while (isspace(*description) && *description != '\0')
		description++;
	
	if (*description == '\0')
		return "pitchwheel had no description";

	gdam_midi_device_add_pitchwheel_name(device, channel, description);
	return NULL;
}

static const char* parse_parameter(GdamMidiDevice* device, const char* args)
{
	int channel, param;
	const char* endp;
	const char* description;
	const char* start_param;
	channel = (int) strtol( (char*) args, (char**) &endp, 10 );
	if (endp == args)
		return "error parsing parameter channel";

	while (isspace(*endp) && *endp != '\0')
		endp++;

	start_param = endp;
	param = (int) strtol( (char*) start_param, (char**) &endp, 10 );
	if (endp == args)
		return "error parsing parameter id";
	
	description = endp;
	while (isspace(*description) && *description != '\0')
		description++;
	gdam_midi_device_add_parameter_name(device,
	                                    channel,
					    param,
					    description);
	return NULL;
}

static const char* parse_sysex(GdamMidiDevice* device, const char* args)
{
	int length, i;
	const char* endp;
	const char* at;
	const char* description;
	guint8* data;
	length = (int) strtol( (char*) args, (char**) &endp, 10 );
	if (endp == args)
		return "error parsing parameter channel";

	while (isspace(*endp) && *endp != '\0')
		endp++;
	at = endp;

	data = alloca(length);

	for (i = 0; i < length; i++) {
		int num;
		num = (int) strtol( (char*) at, (char**) &endp, 16 );
		if (at == endp)
			return "error parsing hex for sysex";
		data[i] = (guint8) num;
		at = endp;
		while (isspace(*at) && *at != '\0')
			at++;
	}
	description = at;
	gdam_midi_device_add_sysex_name(device,
					device->make_id,
					device->model_id,
	                                length,
					data,
					description);
	return NULL;
}

static const char* parse_make(GdamMidiDevice* device, const char* args)
{
	const char* endp;
	int make_id = strtol((char*) args, (char**) &endp, 0);
	if (endp == args) 
		return "error parsing make id";
	while (*endp && isspace(*endp))
		endp++;
	if (*endp != '\0')
		gdam_midi_device_record_make_name(make_id, endp);
	device->make_id = make_id;
	return NULL;
}

static const char* parse_model(GdamMidiDevice* device, const char* args)
{
	const char* endp = NULL;
	int model_id = strtol((char*) args, (char**) &endp, 0);
	if (endp == args) 
		return "error parsing model id";
	while (*endp && isspace(*endp))
		endp++;
	if (*endp != '\0')
		gdam_midi_device_record_model_name(device->make_id,
		                                   model_id,
						   endp);
	device->model_id = model_id;
	return NULL;
}
/*
 * Parse a single midi device patch line, in one of the following
 * formats:
 *
 *     make number name
 *     model number name
 *     pitchwheel 0 name
 *     parameter 0 0 name
 *     sysex length .. ... .. name
 *
 * This function returns an error message, or NULL on success. 
 */
static const char* parse_trap_line(GdamMidiDevice* device, const char* buf)
{
	int cmd_len;
	const char* end_word;
	const char* start_next;
	char* last_cmd;
	static char msg[256];
	while (*buf && isspace(*buf))
		buf++;
	if (*buf == '\0')
		return NULL;
	if (*buf == '#')
		return NULL;
	end_word = buf;
	while (*end_word && ! isspace(*end_word))
		end_word++;
	start_next = end_word;
	while (*start_next && isspace(*start_next))
		start_next++;
	if (*start_next == '\0')
		return "all commands require some parameters";
	cmd_len = (end_word - buf);
	last_cmd = alloca(cmd_len + 1);
	memcpy(last_cmd, buf, cmd_len);
	last_cmd[cmd_len] = '\0';
	if (strcasecmp(last_cmd, "pitchwheel") == 0)
		return parse_pitch_wheel(device, start_next);
	if (strcasecmp(last_cmd, "parameter") == 0)
		return parse_parameter(device, start_next);
	if (strcasecmp(last_cmd, "sysex") == 0)
		return parse_sysex(device, start_next);
	if (strcasecmp(last_cmd, "make") == 0)
		return parse_make(device, start_next);
	if (strcasecmp(last_cmd, "model") == 0)
		return parse_model(device, start_next);

	/* HACK : just so we get a little better error handling,
	          at the expense of totally nonthreadsafe code. */
	if (cmd_len > sizeof(last_cmd) - 1)
		cmd_len = sizeof(last_cmd) - 1;
	memcpy(last_cmd, buf, cmd_len);
	last_cmd[cmd_len] = 0;
	g_snprintf(msg, sizeof(msg),
	           "no command %s found", last_cmd);
	return msg;
}

/* XXX: use gscanner instead. */
gboolean gdam_midi_device_load_map(GdamMidiDevice*    midi_dev,
                                   const char*        db_filename)
{
	FILE* fp;
	char buf[4096];
	int num_errors = 0;
	int lineno = 0;

	if (!gdam_midi_device_has_patchfile(midi_dev, db_filename))
		midi_dev->patch_list = g_slist_prepend(midi_dev->patch_list,
		                                     g_strdup(db_filename));

	g_message("trying to load patch file %s", db_filename);
	fp = fopen(db_filename, "r");
	if (fp == NULL) {
		g_warning("couldn't open %s: %s", db_filename, strerror(errno));
		return FALSE;
	}

	while (fgets(buf, sizeof(buf), fp) != 0) {
		const char* error_message;
		char* nl;
		++lineno;
		nl = strchr(buf, '\n');
		if (nl == NULL) {
			g_warning("line too long or file truncated");
			fclose(fp);
			return FALSE;
		}

		*nl = '\0';	/* chomp */

		error_message = parse_trap_line(midi_dev, buf);
		if (error_message != NULL) {
			g_warning("error parsing trap file %s, line %d, "
				  "`%s': %s, continuing anyway", db_filename,
				  lineno, buf, error_message);
			if (++num_errors > MAX_ERRORS) {
				fclose(fp);
				return FALSE;
			}
		}
	}
	fclose(fp);
	gdam_debug("successfully parsed midi patch database %s", db_filename);
	return TRUE;
}

static void assoc_patch_with_device(const char*   device_name,
                                    const char*   mapfile)
{
	GdamMidiDevice* device = NULL;
	if (midi_devices != NULL)
		device = g_hash_table_lookup(midi_devices, device_name);
	if (device != NULL) {
		if (!gdam_midi_device_has_patchfile(device, mapfile))
			gdam_midi_device_load_map(device, mapfile);
	} else {
		/* No device of that name yet... */
		add_pending_patch(device_name, mapfile);
	}
}

/* Read midi.devices:
 *
 *         /dev/midi???        patch-file-name
 */
static void maybe_initialize_master_patch_table()
{
	FILE* fp;
	char* filename;
	static gboolean has_inited = FALSE;
	char buf[4096];
	int lineno = 0;


	if (has_inited)
		return;

	has_inited = TRUE;

	/* Find ~/.gdam/midi.devices */
        filename = gdam_search_path("midi.devices",
                                    NULL,
		                    GDAM_SEARCH_PATH_FIND_FILE);

	if (filename == NULL) {
		g_message("no midi.devices file -- "
		          "no midi patch information files loaded");
		return;
	}

	fp = fopen(filename, "r");
	if (fp == NULL) {
		g_warning("error opening midi patch description file %s: %s",
			filename, strerror(errno));
		return;
	}
	while (fgets(buf, sizeof(buf), fp)) {
		const char* start_dev_name;
		const char* start_path_fname;
		const char* tmp;
		char* module_name;
		char* path_fname;
		char* dev_name;
		++lineno;
		start_dev_name = buf;
		while (*start_dev_name != '\0' && isspace(*start_dev_name))
			start_dev_name++;
		if (*start_dev_name == '\0'
		 || *start_dev_name == '#')
		 	continue;
		tmp = start_dev_name;
		while (*tmp != '\0' && !isspace(*tmp))
			tmp++;
		if (tmp == start_dev_name || *tmp == '\0') {
			g_warning("too few arguments on line %d", lineno);
			continue;
		}
		dev_name = alloca((tmp - start_dev_name) + 1);
		memcpy(dev_name, start_dev_name, tmp - start_dev_name);
		dev_name[tmp - start_dev_name] = '\0';

		start_path_fname = tmp;
		while (*start_path_fname != '\0' && isspace(*start_path_fname))
			start_path_fname++;
		tmp = start_path_fname;
		while (*tmp != '\0' && !isspace(*tmp))
			tmp++;
		if (tmp == start_path_fname) {
			g_warning("no path on line %d", lineno);
			continue;
		}
		path_fname = alloca((tmp - start_path_fname) + 1);
		memcpy(path_fname, start_path_fname, tmp - start_path_fname);
		path_fname[tmp - start_path_fname] = '\0';

		while (*tmp != '\0' && isspace(*tmp))
			tmp++;

		if (*tmp) {
			const char* end = tmp;
			while (*end != '\0' && !isspace(*end))
				end++;
			module_name = alloca(end + 1 - tmp);
			memcpy(module_name, tmp, end - tmp);
			module_name[end - tmp] = '\0';
		} else
			module_name = NULL;
		assoc_patch_with_device(dev_name, path_fname);
		if (module_name)
			assoc_module_with_device(dev_name, module_name);
	}

	fclose(fp);
}

static void disconnect_hook(GdamMidiHook* hook)
{
	GdamMidiTrap* trap = hook->trap;
	GdamMidiDevice* midi_dev = trap->device;
	trap->hooks = g_slist_remove(trap->hooks, hook);
	g_hash_table_remove(midi_dev->hook_id_table,
	                    GUINT_TO_POINTER(hook->hook_id));
	if (hook->alive_object != NULL) {
		gtk_signal_disconnect(hook->alive_object,
		                      hook->alive_destroy_id);
		hook->alive_object = NULL;
	}
	if (hook->destroy)
		(*hook->destroy)(hook->user_data);
	g_free(hook);
}

void            gdam_midi_device_disconnect
                                     (GdamMidiDevice*    midi_dev,
				      guint              hook_id)
{
	GdamMidiHook* hook;
	g_return_if_fail (midi_dev->hook_id_table != NULL);
	hook = g_hash_table_lookup(midi_dev->hook_id_table,
	                           GUINT_TO_POINTER(hook_id));
	g_return_if_fail(hook != NULL);
	disconnect_hook(hook);
}


typedef struct _DisconnectData DisconnectData;
struct _DisconnectData {
	GdamMidiDevice*    midi_dev;
	GtkObject*         object;
	const char*        hook_name;
	GSList*		   hooks_found;
};

static void
find_hooks_by_hook(gpointer hook_id, gpointer hook_data, gpointer data)
{
	GdamMidiHook* hook = hook_data;
	DisconnectData* disconnect = data;
	if (disconnect->hook_name != NULL
	 && strcmp(hook->hook_name, disconnect->hook_name) != 0)
	 	return;

	/* HACK: alive_object ?!?
	 * this whole `disconnect by object' is a sham...  */
	if (disconnect->object != NULL
	 && hook->alive_object != disconnect->object)
	 	return;

	disconnect->hooks_found
		= g_slist_prepend(disconnect->hooks_found, hook);
}

static void find_all_hooks_for_device(GdamMidiDevice*   device,
                                      DisconnectData*   data)
{
	if (device->trap_table == NULL)
		return;
	g_hash_table_foreach(device->hook_id_table, find_hooks_by_hook, data);
}

static void find_per_device(const char* device_name,
		            GdamMidiDevice* device,
                            DisconnectData* data)
{
	find_all_hooks_for_device(device, data);
}
static void find_all_hooks_for_all(DisconnectData* disconnect_data)
{
	if (midi_devices == NULL)
		return;
	g_hash_table_foreach(midi_devices,
	                     (GHFunc) find_per_device,
			     disconnect_data);
}

void gdam_midi_device_disconnect_by_type (GdamMidiDevice*    midi_dev,
					  GtkObject*         object,
					  const char*        hook_name)
{
	DisconnectData disconnect_data;
	disconnect_data.midi_dev = midi_dev;
	disconnect_data.object = object;
	disconnect_data.hook_name = hook_name;
	disconnect_data.hooks_found = NULL;

	if (midi_dev)
		find_all_hooks_for_device(midi_dev, &disconnect_data);
	else
		find_all_hooks_for_all(&disconnect_data);
	
	g_slist_foreach(disconnect_data.hooks_found,
	                (GFunc) disconnect_hook,
			NULL /* unused */);
	g_slist_free(disconnect_data.hooks_found);
}

typedef struct _TrapData TrapData;
struct _TrapData {
	GdamMidiTrapIterator iter_func;
	gpointer             iter_data;
};
	
static void iterate_devices(const char*     device_name,
                            GdamMidiDevice* device,
			    TrapData*       trap_data)
{
	gdam_midi_device_iterate_traps(device,
	                               trap_data->iter_func,
	                               trap_data->iter_data);
}
static void iterate_trap   (GdamMidiTrap*   trap,
                            GdamMidiTrap*   same_as_trap,
			    TrapData*       trap_data)
{
	(*trap_data->iter_func)(trap, trap_data->iter_data);
}

void gdam_midi_device_iterate_traps     (GdamMidiDevice*      device,
					 GdamMidiTrapIterator iter_func,
					 gpointer             iter_data)
{
	TrapData trap_data;
	trap_data.iter_func = iter_func;
	trap_data.iter_data = iter_data;
	if (device == NULL) {
		if (midi_devices != NULL)
			g_hash_table_foreach(midi_devices,
					     (GHFunc) iterate_devices,
					     &trap_data);
		return;
	}

	g_hash_table_foreach(device->trap_table,
	                     (GHFunc) iterate_trap,
			     &trap_data);
}
