/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamresample.h"
#include "gdamdispatch.h"
#include <gdam/gdam-protocol.h>
#include <gdam/gdamarg.h>
#include <stdio.h>


static int         gdam_resample_read             (GdamSource*	    resample,
                                       	           GdamSample*	    buf,
				       	           int		    n);
static void        gdam_resample_miss_data        (GdamSource*	    source,
                                       	           int              amount);
static gboolean    gdam_resample_seek             (GdamSource*      source,
                                                   GdamTime         offset,
		                                   int              whence);
static GdamTime    gdam_resample_get_position     (GdamSource*      source);

static void        gdam_resample_get_latency_info (GdamSource*      source,
					           GdamSource*      subsource,
					           GdamLatencyInfo* latency);

static void        gdam_resample_class_init       (GdamFilterClass* klass);
static void        gdam_resample_init             (GdamResample*    resample);

static GtkObjectClass* parent_class;

GtkType gdam_resample_get_type()
{
	static GtkType resample_type = 0;
	if (!resample_type) {
		static const GtkTypeInfo resample_info = {
			"GdamResample",
			sizeof(GdamResample),
			sizeof(GdamFilterClass),
			(GtkClassInitFunc) gdam_resample_class_init,
			(GtkObjectInitFunc) gdam_resample_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		parent_class = gtk_type_class(GDAM_TYPE_FILTER);
		resample_type = gtk_type_unique(GDAM_TYPE_FILTER, 
		                                &resample_info);
	}
	return resample_type;
}
static void gdam_resample_arg_set_rate(GtkObject* object, GdamArg* arg)
{
	GdamResample* resample = GDAM_RESAMPLE(object);
	guint64 rate = (guint64)(GDAM_ARG_DOUBLE(arg) * 65536.0 * 65536.0);
	resample->rate = rate;
	resample->last_skipped = 0;
}
static void gdam_resample_arg_get_rate(GtkObject* object, GdamArg* arg)
{
	GdamResample* resample = GDAM_RESAMPLE(object);
	GDAM_ARG_DOUBLE(arg) = (double)resample->rate / (65536.0 * 65536.0);
}
static void gdam_resample_arg_set_interpolating(GtkObject* object, GdamArg* arg)
{
	GdamResample* resample = GDAM_RESAMPLE(object);
	resample->interpolating = GDAM_ARG_INT(arg);
}
static void gdam_resample_arg_get_interpolating(GtkObject* object, GdamArg* arg)
{
	GdamResample* resample = GDAM_RESAMPLE(object);
	GDAM_ARG_INT(arg) = resample->interpolating;
}

static void gdam_resample_get_latency_info(GdamSource*      source,
					   GdamSource*      subsource,
					   GdamLatencyInfo* latency)
{
	latency->latency = 0;
	g_return_if_fail(GDAM_RESAMPLE(source) != NULL);
	latency->resample = (GdamFixed) (((GdamResample*)source)->rate >> 16);
}


static void gdam_resample_class_init(GdamFilterClass* klass)
{
	GdamSourceClass* source_class = (GdamSourceClass*)klass;
	source_class->read = gdam_resample_read;
	source_class->miss_data = gdam_resample_miss_data;
	source_class->seek = gdam_resample_seek;
	source_class->get_position = gdam_resample_get_position;
	source_class->get_latency_info = gdam_resample_get_latency_info;
	gdam_arg_register(GTK_OBJECT_CLASS(klass),
	                  "rate",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_resample_arg_set_rate,
			  gdam_resample_arg_get_rate,
			  GDAM_ARG_VITAL);
	gdam_arg_register(GTK_OBJECT_CLASS(klass),
	                  "interpolating",
			  GDAM_ARG_TYPE_INT,
			  gdam_resample_arg_set_interpolating,
			  gdam_resample_arg_get_interpolating,
			  GDAM_ARG_VITAL);
}

static void gdam_resample_init(GdamResample* resample)
{
	resample->accum = 0;
	resample->last_sample.s[0] = resample->last_sample.s[1] = 0;
	resample->rate = (1LL<<32);
}


static int gdam_resample_read(GdamSource* source, GdamSample* buf, int n)
{
	GdamResample* resample = GDAM_RESAMPLE(source);
	GdamFilter* filter = GDAM_FILTER(source);
	GdamSource* filtered;
	int num_read, rv;
	guint64 rate;
	guint accum;
	GdamSample last_sample;
	GdamSample* at_sample;     /* will point inside the buffer `tmp' */
	/* How many samples can we get without overfilling buf 
	   and last_sample? */
	int max_read;
	int num_output;
	int leftover_input;
	int last_skipped;

	GdamSample* tmp;
	GdamSample* out;

	g_return_val_if_fail(resample != NULL, 0);
	filtered = filter->filtered;
	if (!filtered)
		return 0;
	rate = resample->rate;
	accum = resample->accum;

	if (rate == (1LL<<32))
		return gdam_source_read(filtered, buf, n);

	/* Determine the constraints on how much data to read,
	 * we subtract 1 because the leftover counts as a sample. */
	if (resample->last_skipped > 0) {
		last_skipped = resample->last_skipped;
	 	if (last_skipped > (rate>>32)) {
			g_warning("cropping last_skipped [was %d]",
				last_skipped);
			last_skipped = (rate>>32) - 1;
		}
	} else
		last_skipped = 0;
	max_read = (int) ((resample->rate * n + accum) >> 32) - last_skipped;

	if (max_read <= 0) {
		tmp = NULL;
		num_read = 0;
	} else {
		/* Read the data. */
		tmp = (GdamSample*) alloca(sizeof(GdamSample) * max_read);
		num_read = gdam_source_read(filtered, tmp, max_read);

		if (num_read <= 0) {
			/* Just discard fractional ending data... who cares? */
			return num_read;
		}
	}

	/* Prevent division by zero and overflows... */
	if (num_read >= resample->rate) {
		num_output = n;
	} else {
		num_output = ((((gint64)(num_read + 1 + last_skipped)
		                           << 32)) - accum - 1)
		           / resample->rate;
		if (num_output > n) {
			num_output = n;
		}
	}

	if (num_read == max_read && num_output != n) {
		g_warning("n=%d, num_output=%d, num_read=max_read=%d",
			n,num_output,max_read);
	}

	/* The use of `tmp - 1' is to denote that we are
	 * outputting `last_sample' right now. */
	at_sample = tmp - 1 - last_skipped;
	out = buf;
	rv = num_output;
	last_sample = resample->last_sample;

	/* Rescale the data. */
	{
		guint32 fraction = (guint32)rate;
		guint32 whole = (guint32)(rate>>32);
		while (num_output-- > 0) {
			guint32 new_accum = accum + fraction;
			at_sample += whole;
			if (new_accum < accum) {
				last_sample = *(++at_sample);
			} else if (whole)
				last_sample = *at_sample;
			accum = new_accum;

			*out++ = last_sample;
		}
	}

	resample->last_sample = last_sample;
	resample->accum = accum;

	/* Not maintaining the following condition will lead
	 * to pops and so forth, as the source above us
	 * will think we are blocking... */
	if (num_read == max_read)
		g_return_val_if_fail(rv == n, rv);

	if (tmp != NULL) {
		leftover_input = tmp + num_read - 1 - at_sample;
		resample->last_skipped = leftover_input;
		g_return_val_if_fail(leftover_input >= 0, rv);
	} else {
		resample->last_skipped = 0;
	}
	g_return_val_if_fail(out - buf == rv, rv);

	return rv;
}
static void        gdam_resample_miss_data        (GdamSource*	    source,
                                       	           int              n)
{
	GdamResample* resample = GDAM_RESAMPLE(source);
	GdamSource* filtered = GDAM_FILTER(source)->filtered;
	if (filtered != NULL) {
		int submiss;
		submiss = (int) ((resample->accum + resample->rate * n) >> 32);
		gdam_source_miss_data(filtered, submiss);
	}
	resample->accum += (guint32)(resample->rate * n);
	return;
}

static gboolean
gdam_resample_seek(GdamSource* source,
                   GdamTime    offset,
		   int         whence)
{
	GdamResample* resample = GDAM_RESAMPLE(source);
	GdamFilter* filter = GDAM_FILTER(source);
	g_return_val_if_fail(resample != NULL, FALSE);
	g_return_val_if_fail(filter != NULL, FALSE);

	if (filter->filtered == NULL) {
		gdam_debug("seek attempted on resample with no subsource");
		return FALSE;
	}

	/* XXX: not safe to 64-bits at all!!! */
	offset *= resample->rate;
	offset >>= 32;

	return gdam_source_seek(filter->filtered, offset, whence);
}

static GdamTime gdam_resample_get_position(GdamSource* source)
{
	GdamResample* resample = GDAM_RESAMPLE(source);
	GdamFilter* filter = GDAM_FILTER(source);
	gdouble position;
	if (filter->filtered == NULL) {
		gdam_debug("tried to get position of an empty resample");
		return -1;
	}
	position = gdam_source_get_position(filter->filtered);
	if (position == -1)
		return -1;
	
	position *= (1LL<<32);
	if (resample->rate == 0) {
		g_warning("resample rate was 0");
		return -1;
	}
	position /= resample->rate;
	return (gint64) position;
}


