/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Mike Radford, Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
	mrad@blorf.com <Mike Radford>
*/
#include <gdam/gdamsource.h>
#include "gdamcompressingmix.h"
#include <gdam/gdamdispatch.h>
#include <gdam/gdamnamespace.h>
#include <gdam/gdam-protocol.h>
#include <assert.h>
#include <stdio.h>

#include "limiter-table.inc"

#define DEFAULT_MIX_BUFFER	2048

/* Please look at the casts to this function below before
 * changing its prototype. */
typedef void (*SampleTransformFunc)(int*              dest,
                                    const GdamSample* src,
				    int               num,
				    gpointer          data);
	

static void        gdam_compressing_mix_class_init       (GdamSourceClass* klass);
static void        gdam_compressing_mix_init             (GdamCompressingMix*         mix);
static void        gdam_compressing_mix_destroy          (GdamSource*      source);
static GdamBuffer* gdam_compressing_mix_interpret_command(GdamSource*      source,
					      GdamBuffer*      command,
					      gboolean*        was_ok);
static gboolean    gdam_compressing_mix_is_ready_to_read (GdamSource*      source);
static int	   gdam_compressing_mix_read             (GdamSource*      input,
					      GdamSample*      samples,
					      int              max_read);
static GdamTime    gdam_compressing_mix_get_position     (GdamSource*      source);
static void        gdam_compressing_mix_get_latency_info (GdamSource*      source_mix,
                                              GdamSource*      sub,
				              GdamLatencyInfo* info);
static void        gdam_compressing_mix_compress(GdamCompressingMix* mix,
				                 GdamSample*         compressed,
	                                         int*                uncompressed,
				                 int                 num_samples);
static void        gdam_compressing_mix_miss_data        (GdamSource*      input,
					      int              lost);
static gboolean    gdam_compressing_mix_detach_child     (GdamSource*      mix,
                                              GdamSource*      sub_source);

static void        delete_element            (GdamCompressingMix* mix,
                                              GdamCompressingMixElement* element);

/* TODO:  on each read, sort the elements by the amount of
 *        blank space they have at the beginning... we should
 *        gdam_sample_zero only if there are no sources ready to read... */

static GtkObjectClass* parent_class = NULL;


static void gdam_compressing_mix_init(GdamCompressingMix* mix)
{
	mix->first_source      = 0;
	mix->last_source       = 0;
	mix->buffer_size       = DEFAULT_MIX_BUFFER;
	mix->buffer            = g_new(int, mix->buffer_size*2);
	mix->buffer_data_start = 0;
	mix->tmp_buf           = g_new(GdamSample, mix->buffer_size);
	mix->samples_output    = 0;
}

#ifdef gdam_sample_copy
#undef gdam_sample_copy
#endif

static void gdam_compressing_mix_set_mixed(GtkObject* object, GdamArg* arg)
{
	GdamCompressingMix* mix = (GdamCompressingMix*)object;
	gint64* array = GDAM_ARG_INT64_ARRAY(arg).array;
	int num_els = GDAM_ARG_INT64_ARRAY(arg).length, i;
	g_return_if_fail(num_els % 2 == 0);
	num_els /= 2;

	while (mix->first_source)
		delete_element(mix, mix->first_source);
	
	for (i = 0; i < num_els; i++) {
		gint64 position = (int)array[2*i+0];
		int source_id = (int)array[2*i+1];
		GdamSource* subsource = gdam_namespace_local_source(source_id);
		if (subsource == NULL) {
			gdam_debug("no source of id %d "
					"(needed in setting mixed)");
			continue;
		}
		gdam_compressing_mix_connect(mix, subsource, position);
	}
}

static void gdam_compressing_mix_get_mixed(GtkObject* object, GdamArg* arg)
{
	GdamCompressingMix* mix = GDAM_COMPRESSING_MIX(object);
	GdamCompressingMixElement* mix_element = mix->first_source;
	int num_subs = gdam_compressing_mix_num_sources(mix);
	gint64* rv = g_new(gint64, num_subs * 2);
	int i;
	for (i = 0; i < num_subs; i++) {
		rv[i*2 + 0] = mix_element->read_to;
		rv[i*2 + 1] = mix_element->sub_source->source_id;

		mix_element = mix_element->next;
	}
	GDAM_ARG_INT64_ARRAY(arg).length = num_subs * 2;
	GDAM_ARG_INT64_ARRAY(arg).array = rv;
}
static void gdam_compressing_mix_set_terminates(GtkObject* object, GdamArg* arg)
{
	GDAM_COMPRESSING_MIX(object)->terminates = GDAM_ARG_INT(arg) ? 1 : 0;
}
static void gdam_compressing_mix_get_terminates(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = GDAM_COMPRESSING_MIX(object)->terminates;
}


static void gdam_compressing_mix_class_init(GdamSourceClass* klass)
{
	GtkObjectClass* object_class = (GtkObjectClass*)klass;
	GdamCompressingMixClass* mix_class = (GdamCompressingMixClass*)klass;
	if (!parent_class)
		parent_class = gtk_type_class(GDAM_TYPE_SOURCE);
	klass->read = gdam_compressing_mix_read;
	klass->miss_data = gdam_compressing_mix_miss_data;
	klass->interpret_command = gdam_compressing_mix_interpret_command;
	klass->is_ready_to_read = gdam_compressing_mix_is_ready_to_read;
	klass->detach_child = gdam_compressing_mix_detach_child;
	klass->get_latency_info = gdam_compressing_mix_get_latency_info;
	klass->get_position = gdam_compressing_mix_get_position;
	mix_class->compress = gdam_compressing_mix_compress;
	object_class->destroy = (void(*)(GtkObject*)) gdam_compressing_mix_destroy;

	gdam_arg_register(object_class,
	                  "mixed",
			  GDAM_ARG_TYPE_INT64_ARRAY,
			  gdam_compressing_mix_set_mixed,
			  gdam_compressing_mix_get_mixed,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "terminates",
			  GDAM_ARG_TYPE_INT,
			  gdam_compressing_mix_set_terminates,
			  gdam_compressing_mix_get_terminates,
			  GDAM_ARG_VITAL);
}

static void delete_element(GdamCompressingMix* mix, GdamCompressingMixElement* element)
{
	gdam_debug("mix_delete_element: element=%p; source=%p",
		element, element->sub_source);
	element->sub_source->parent = NULL;
	if (element->next)
		element->next->prev = element->prev;
	else
		mix->last_source = element->prev;
	if (element->prev)
		element->prev->next = element->next;
	else
		mix->first_source = element->next;
	gtk_object_unref((GtkObject*)(element->sub_source));
	g_free(element);
}

/* gdam_compressing_mix_subsource read calls ready_to_read on subsources,
   so it can deal with any situation... in fact, that unifies the places
   we have to deal with it...

   this might not be great if algorithms ever called gdam_compressing_mix_is_ready_to_read
   explorationally, but it suffices for now. */
static gboolean gdam_compressing_mix_is_ready_to_read(GdamSource* source)
{
	return TRUE;

	/*GdamCompressingMixElement* element;

	for (element = mix->first_source;
	     element != NULL;
	     element = element->next) {
	     	if (element->started <= mix->samples_output
		 && element->read_to > mix->samples_output)
		 	return TRUE;
		if (element->started > mix->samples_output)
			continue;
		
	     ...

	}*/
}
	
static inline void init_func(int*              dest,
                             const GdamSample* src,
	                     int               num,
			     gpointer          data)
{
	while (num-- > 0) {
		*dest++ = src->s[0];
		*dest++ = src->s[1];
		src++;
	}
}

static inline void add_func(int*              dest,
                            const GdamSample* src,
	                    int               num,
			    gpointer          data)
{
	while (num-- > 0) {
		*dest++ += src->s[0];
		*dest++ += src->s[1];
		src++;
	}
}

/* Read from an element, returning whether or not to keep the source
 * around. */
static gboolean gdam_compressing_mix_subsource_read(GdamCompressingMix* mix, GdamCompressingMixElement* element)
{
	int max_read, r;
	int init_buffer_index;
	int num_clear;
	gint64 max_end_buf = mix->samples_output + mix->buffer_size;
	GdamCompressingMixClass* mix_class;
	gpointer element_data;


	if (max_end_buf <= element->read_to)
		return TRUE;
	max_read = (max_end_buf - element->read_to);
	g_assert(max_read <= mix->buffer_size);
	r = gdam_source_read(element->sub_source, mix->tmp_buf, max_read);
	if (r <= 0) {
		return FALSE;
	}
	num_clear = (int)(element->read_to + r - mix->cleared_to);
	if (num_clear < 0)
		num_clear = 0;
	else {
		if (num_clear > r) {
			num_clear = r;
			r = 0;
		} else {
			r -= num_clear;
		}
	}

	init_buffer_index = (int)(element->read_to % mix->buffer_size);

	element_data = element->source_info;

	mix_class = (GdamCompressingMixClass*) ((GtkObject*)mix)->klass;
	if (r > 0) {
		//add_func = mix_class->augment;
		if (init_buffer_index + r > mix->buffer_size) {
			int second_part = mix->buffer_size - init_buffer_index;
			int first_part = r - second_part;
			assert(second_part > 0 && first_part > 0);
			add_func(mix->buffer + init_buffer_index * 2,
					  mix->tmp_buf,
					  second_part,
					  element_data);
			add_func(mix->buffer,
					  mix->tmp_buf + second_part,
					  first_part,
					  element_data);
		} else {
			add_func(mix->buffer + init_buffer_index * 2,
				 mix->tmp_buf,
				 r,
				 element_data);
		}
	}
	element->read_to += r;
	if (num_clear > 0) {
		init_buffer_index += r;
		init_buffer_index %= mix->buffer_size;
		//init_func = mix_class->initial_copy;

		if (init_buffer_index + num_clear > mix->buffer_size) {
			int second_part = mix->buffer_size - init_buffer_index;
			int first_part = num_clear - second_part;
			assert(second_part > 0 && first_part > 0);
			init_func(mix->buffer + init_buffer_index * 2,
					  mix->tmp_buf + r,
					  second_part,
					  element_data);
			init_func(mix->buffer,
					  mix->tmp_buf + second_part + r,
					  first_part,
					  element_data);
		} else {
			init_func(mix->buffer + init_buffer_index * 2,
				 mix->tmp_buf + r,
				 num_clear,
				 element_data);
		}
		mix->cleared_to += num_clear;
		element->read_to += num_clear;
	}
	return TRUE;
}
/*
static inline void
gdam_sample_copy_and_zero(GdamSample* dst, GdamSample* src, int n)
{
	gdam_sample_copy(dst, src, n);
	gdam_sample_zero(src, n);
}
*/

static int gdam_compressing_mix_read(GdamSource* source, GdamSample* buf, int n)
{
	GdamCompressingMix* mix = (GdamCompressingMix*)source;
	GdamCompressingMixElement* element, *next_element = 0;
	void (*compress)(GdamCompressingMix*, GdamSample*, int*, int);
	int to_read = n;
	GdamTime start_buffer_time = mix->samples_output;
	GdamTime end_buffer_time = start_buffer_time + mix->buffer_size;
	GdamTime end_read_time;

	compress = GDAM_COMPRESSING_MIX_CLASS(GTK_OBJECT(source)->klass)
							->compress;

	/* Search through the sources and see if we can
	 * read on sources we want to read.  */
	for (element = mix->first_source;
	     element != NULL;
	     element = next_element) {
		element->flags = 0;
		next_element = element->next;
		/* if the element is too far in the future to read... */
		if (element->read_to >= end_buffer_time) {
			element->flags |= MIX_ELEMENT_IN_FUTURE;
			continue;
		}

	     	/* if we could read more, check if we want to read more. */
		if (gdam_source_is_ready_to_read(element->sub_source))
			if (!gdam_compressing_mix_subsource_read(mix, element)) {
				gdam_debug("gdam_compressing_mix_read: "
					"calling delete_element");
				delete_element(mix, element);
			}
	}

	/* Compute the amount to return.  (If sources were unable
	 * to read, and have no data buffered they will have the
	 * `miss_data' callback called.) */
	for (element = mix->first_source;
	     element != NULL;
	     element = element->next) {
		int data_available;
	     	if (element->flags & MIX_ELEMENT_IN_FUTURE)
			continue;
		data_available = (int)(element->read_to - start_buffer_time);
		if (data_available > 0
		 && data_available < to_read)
			to_read = data_available;
	}

	/* Call the miss_data virtuals, as needed. */
	end_read_time = start_buffer_time + to_read;
	for (element = mix->first_source;
	     element != NULL;
	     element = element->next) {
		int missed = (int)(end_read_time - element->read_to);
	     	if (missed > 0) {
			gdam_source_miss_data(element->sub_source, missed);
			element->read_to = end_read_time;
		}
	}

	if (mix->cleared_to < end_read_time) {
		int num_to_clear = (int)(end_read_time - mix->cleared_to);
		if (num_to_clear > mix->buffer_size) {
			gdam_sample_zero(mix->buffer, mix->buffer_size);
		} else if (num_to_clear > 0) {
			int start_clear_index = (int)(mix->cleared_to
			                            % mix->buffer_size);
			int p1 = mix->buffer_size - start_clear_index;
			int* buf_start = mix->buffer + start_clear_index*2;
			if (p1 > num_to_clear) {
				gdam_sample_zero(buf_start, num_to_clear);
				mix->cleared_to += num_to_clear;
				mix->buffer_data_start = 0;
			} else {
				int p2 = num_to_clear - p1;
				gdam_sample_zero(buf_start, p1);
				gdam_sample_zero(mix->buffer, p2);
				mix->cleared_to += num_to_clear;
				mix->buffer_data_start = p2;
			}
		}
		mix->cleared_to = end_read_time;
	}
	/* Copy the data out of the buffer. */
	{
		int p1 = mix->buffer_size - mix->buffer_data_start;
		int* buf_start = mix->buffer + 2*mix->buffer_data_start;
		g_assert(to_read > 0);
		if (p1 > to_read) {
			compress(mix, buf, buf_start, to_read);
			mix->buffer_data_start += to_read;
		} else {
			int p2 = to_read - p1;
			compress(mix, buf, buf_start, p1);
			compress(mix, buf + p1, mix->buffer, p2);
			mix->buffer_data_start = p2;
		}
	}
	mix->samples_output += to_read;
	return to_read;
}

static void gdam_compressing_mix_get_latency_info(GdamSource* source_mix,
                                      GdamSource* sub,
				      GdamLatencyInfo* info)
{
	GdamCompressingMixElement* element;
	GdamCompressingMix* mix = GDAM_COMPRESSING_MIX(source_mix);
	info->resample = GdamFixed_ONE;
	g_return_if_fail(mix != NULL);
	for (element = mix->first_source;
	     element != NULL;
	     element = element->next) {
	     	if (element->sub_source == sub) {
			int delta;
			delta = (int)(element->read_to - mix->samples_output);
			info->latency = delta;
			return;
		}
	}
	info->latency = 0;
	return;
}

/* TODO: fix these two functions (eliminate destruct) */
void mix_destruct(GdamSource* source)
{
	GdamCompressingMix* mix = (GdamCompressingMix*)source;
	while (mix->first_source)
		delete_element(mix, mix->first_source);
	g_free(mix->buffer);
	mix->buffer = 0;
	g_free(mix->tmp_buf);
	mix->tmp_buf = 0;
}

static void gdam_compressing_mix_destroy(GdamSource* source)
{
	GdamCompressingMix* mix = GDAM_COMPRESSING_MIX(source);
	while (mix->first_source)
		gdam_source_detach_child(source, mix->first_source->sub_source);
	g_free(mix->buffer);
	mix->buffer = NULL;
}

void gdam_compressing_mix_connect(GdamCompressingMix* mix, GdamSource* add, gint64 when)
{
	GdamCompressingMixElement* element;
	GdamCompressingMixElement* new_element;
	GtkObject* as_object = GTK_OBJECT(add);
	if (when == -1)
		when = mix->samples_output;

	/* This increases the ref count of add. */
	gdam_source_orphan(add);

	for (element = mix->first_source;
	     element && element->read_to < when;
	     element = element->next)
	     	;

	new_element = g_new(GdamCompressingMixElement, 1);
	new_element->next = element;
	new_element->prev = element ? element->prev : mix->last_source;

	if (new_element->prev)
		new_element->prev->next = new_element;
	else
		mix->first_source = new_element;
	if (new_element->next)
		new_element->next->prev = new_element;
	else
		mix->last_source = new_element;
	gdam_debug("gdam_compressing_mix_connect: source %p (element %p) time %lld",
		add, new_element, when);
	new_element->sub_source = add;
	gtk_object_sink(as_object);
	new_element->read_to = when;
	add->parent = (GdamSource*)mix;
}

static GdamTime    gdam_compressing_mix_get_position     (GdamSource*      source)
{
	return GDAM_COMPRESSING_MIX(source)->samples_output;
}
static void gdam_compressing_mix_miss_data(GdamSource* source, int lost)
{
	GdamCompressingMix* mix = GDAM_COMPRESSING_MIX(source);
	GdamTime new_read_to = mix->samples_output + lost;
	GdamCompressingMixElement* element;
	for (element = mix->first_source;
	     element;
	     element = element->next) {
	     	if (element->read_to < new_read_to) {
			int delta = (int)(new_read_to - element->read_to);
			gdam_source_miss_data(element->sub_source, delta);
			element->read_to = new_read_to;
	     	}
	}
	mix->samples_output = new_read_to;
}

static gboolean gdam_compressing_mix_detach_child(GdamSource* super_source,
                                      GdamSource* sub_source)
{
	GdamCompressingMix* mix = GDAM_COMPRESSING_MIX(super_source);
	GdamCompressingMixElement* element;
	g_return_val_if_fail(mix != NULL, FALSE);
	for (element = mix->first_source;
	     element && element->sub_source != sub_source;
	     element = element->next)
	     	;
	g_return_val_if_fail(element != NULL, FALSE);
	if (!element)
		return FALSE;
	gdam_debug("gdam_compressing_mix_detach_child: calling delete_element");
	delete_element(mix, element);
	return TRUE;
}

static GdamBuffer* gdam_compressing_mix_interpret_command(GdamSource*      source,
					      GdamBuffer*      command,
					      gboolean*        was_ok)
{
	GdamSourceClass* par_source_class = (GdamSourceClass*)parent_class;
	int type;
	if (gdam_buffer_peek_int(command, &type)) {
		if (type == GDAM_CFG_MIX_ADD) {
			GdamCompressingMix* mix;
			GdamSource* subsource;
			int source_id;
			GdamTime when;
			mix = GDAM_COMPRESSING_MIX(source);
			gdam_buffer_parse_int(command, &type);
			if (!gdam_buffer_parse_int(command, &source_id)) {
			 	*was_ok = FALSE;
				gdam_debug("mix_add: no source_id ");
				return NULL;
			}
			subsource = gdam_namespace_local_source(source_id);
			if (!subsource) {
			 	*was_ok = FALSE;
				gdam_debug("mix_add: bad source id %d",
					source_id);
				return NULL;
			}

			if (!gdam_buffer_parse_time(command, &when)) {
				*was_ok = FALSE;
				gdam_debug("mix_add: no time (too short)");
				return NULL;
			}
			gdam_compressing_mix_connect(mix, subsource, when);
			*was_ok = TRUE;
			return NULL;
		}
	}
	return par_source_class->interpret_command(source, command, was_ok);
}

int gdam_compressing_mix_num_sources(GdamCompressingMix* mix)
{
	GdamCompressingMixElement* element;
	int count = 0;
	for (element = mix->first_source;
	     element;
	     element = element->next)
	     	count++;
	return count;
}

static inline gint16 limit_s16(int in)
{
	int sign;
	if (in < 0) {
		in = -in;
		sign = -1;
	} else {
		sign = 1;
	}

	if (in >= 181705)
		return sign * 32767;

	{
	int lo = 0;
	int hi = 7811;
	int mid;

	if (in < 32768) {
		/* use table of (input-output) vs. input for lower range */
		while (lo < hi) {
			mid = (lo + hi) / 2;
			if (limit_low_s16[mid] < in)
				lo = mid + 1;
			else
				hi = mid;
		}
		return (gint16) sign *(in - mid);
	} else {
		/* use table of output vs. input for upper range */
		while (lo < hi) {
			mid = (lo + hi) / 2;
			if (limit_high_s16[mid] < in)
				lo = mid + 1;
			else
				hi = mid;
		}
		return (gint16) sign *(mid + 24955);
	}
	}
}

static void gdam_compressing_mix_compress(GdamCompressingMix* mix,
				          GdamSample*         compressed,
	                                  int*                uncompressed,
				          int                 num_samples)
{
	int i;
	for (i = 0; i < num_samples; ++i) {
		compressed[i].s[0] = limit_s16(uncompressed[2*i]);
		compressed[i].s[1] = limit_s16(uncompressed[2*i+1]);
	}
}

GtkType gdam_compressing_mix_get_type()
{
	static GtkType mix_type = 0;
	if (!mix_type) {
		static const GtkTypeInfo mix_info = {
			"GdamCompressingMix",
			sizeof(GdamCompressingMix),
			sizeof(GdamCompressingMixClass),
			(GtkClassInitFunc) gdam_compressing_mix_class_init,
			(GtkObjectInitFunc) gdam_compressing_mix_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		parent_class = gtk_type_class(GDAM_TYPE_SOURCE);
		mix_type = gtk_type_unique(GDAM_TYPE_SOURCE, &mix_info);
	}
	return mix_type;
}
